import re

P_VALUE_STRINGS = ('P>|z|', 'P-value', 'P>|t|', 'P>|F|')
KNOWN_ABBREVIATIONS = ('std', 'BMI', 'Std.', 'Std', 'Err.', 'Avg.', 'Coef.', 'SD', 'SE', 'CI') \
                      + P_VALUE_STRINGS


def contains_both_letter_and_numbers(name: str) -> bool:
    """
    Check if the name contains both letters and numbers.
    """
    return any(char.isalpha() for char in name) and any(char.isdigit() for char in name)


def is_unknown_abbreviation(name: str) -> bool:
    """
    Check if the name is abbreviated.
    """
    if not isinstance(name, str):
        return False

    if len(name) == 0:
        return False

    if name.isnumeric():
        return False

    if name in KNOWN_ABBREVIATIONS:
        return False

    if len(name) <= 2:
        return True

    # if contains underscore, treat as abbreviation
    if '_' in name:
        return True

    for abbreviation in KNOWN_ABBREVIATIONS:
        if abbreviation.endswith('.'):
            pattern = r'\b' + re.escape(abbreviation)
        else:
            pattern = r'\b' + re.escape(abbreviation) + r'\b'
        name = re.sub(pattern, '', name)

    # if there are no letters left, it is not an abbreviation
    if not any(char.isalpha() for char in name):
        return False

    words = re.split(pattern=r'[-_ =(),]', string=name)
    if any(contains_both_letter_and_numbers(word) for word in words):
        return True

    # if there are over 3 words, it is not an abbreviation:
    if len(re.split(pattern=r' ', string=name)) >= 3:
        return False

    if '.' in name or ':' in name or '_' in name:
        return True
    words = re.split(pattern=r'[-_ ]', string=name)
    words = [word for word in words if word != '']
    if all((word.islower() or word.istitle()) or word.isnumeric()
           or (word[:-1].isnumeric() and word[-1] == '%') for word in words):
        return False
    return True



import re
import textwrap
from typing import Optional, Union, Tuple, Dict

ArgsOrKwargs = Union[Tuple[str], Dict[str, str]]


def format_with_args_or_kwargs(text: str, args_or_kwargs: ArgsOrKwargs) -> str:
    """
    Return the text formatted with the given args or kwargs.
    """
    if isinstance(args_or_kwargs, tuple):
        return text.format(*args_or_kwargs)
    else:
        return text.format(**args_or_kwargs)


def dedent_triple_quote_str(s: str, indent: int = 0) -> str:
    """
    Format a triple-quote string to remove extra indentation and leading newline.
    """
    if s.startswith('\n'):
        s = s[1:]
    s = textwrap.dedent(s)
    s = s.replace('\t\n', '')  # allows continuing lines in triple-quote strings by ending with \t
    if indent > 0:
        s = textwrap.indent(s, ' ' * indent)
    return s


def wrap_string(input_string, width: Optional[int] = 40, indent: int = 0, new_line_indent: bool = False):
    """
    Add linebreaks to wrap a long string.
    """
    # split input string into lines
    lines = input_string.split('\n')
    wrapped_lines = []

    # wrap each line individually
    for line in lines:
        if width is None:
            wrapped_line = line
        else:
            wrapped_line = textwrap.fill(line, width=width).replace('\n', '\n\t' if new_line_indent else '\n')
        wrapped_lines.append(wrapped_line)

    # join wrapped lines back together with preserved line breaks
    wrapped_string = "\n".join(wrapped_lines)

    # add indent to each line if specified
    if indent > 0:
        wrapped_string = textwrap.indent(wrapped_string, ' ' * indent)

    return wrapped_string


def wrap_python_code(code, width=70):
    wrapped_lines = []
    for line in code.split('\n'):
        stripped_line = line.strip()
        # Wrap comments
        if stripped_line.startswith("#"):
            # Preserve the leading whitespace
            leading_whitespace = line[:line.index("#")]
            wrapped_comment = textwrap.fill(stripped_line, width=width,
                                            initial_indent=leading_whitespace,
                                            subsequent_indent=leading_whitespace + "# ",
                                            break_long_words=False,
                                            break_on_hyphens=False)
            wrapped_lines.extend(wrapped_comment.split('\n'))
        # Wrap non-empty lines that are not comments
        elif stripped_line:
            wrapped_line = textwrap.wrap(line, width=width,
                                         break_long_words=False,
                                         break_on_hyphens=False)
            for i, part in enumerate(wrapped_line[:-1]):
                wrapped_lines.append(part + " \\")
            # Align the continuation lines with the original line's indentation
            continuation_indent = len(line) - len(line.lstrip())
            wrapped_lines.append(" " * continuation_indent + wrapped_line[-1])
        else:
            wrapped_lines.append(line)
    return "\n".join(wrapped_lines)


def format_str_by_direct_replace(text: str, replacements: dict):
    """
    Replace all occurrences of the keys in replacements with their corresponding values.
    """
    for key, value in replacements.items():
        text = text.replace('{' + key + '}', str(value))
    return text


def wrap_as_block(text: str, header: str = '') -> str:
    """
    Wrap text with triple quotes.
    """
    text = text.strip('\n')
    return f'```{header}\n{text}\n```'


def forgiving_format(string, *args, **kwargs):
    """
    A forgiving version of str.format() that returns the original string if there are no matching arguments.
    """

    # Regular expression pattern to match placeholders in the string
    pattern = re.compile(pattern=r'\{\{.*?\}\}|\{.*?\}')  # {{var}} or {var}

    def substitute(match):
        nonlocal args
        match = match.group()
        if match[:2] == '{{' and match[-2:] == '}}':
            return match[1:-1]
        if match == '{}':
            if len(args) > 0:
                replace_with = str(args[0])
                args = args[1:]
            else:
                replace_with = '{}'
        else:
            key = match[1:-1]
            if key in kwargs:
                replace_with = str(kwargs[key])
            else:
                replace_with = match
        return replace_with

    return re.sub(pattern, substitute, string)


def short_repr(var):
    """
    Return a short representation of the given variable.
    """
    rep = repr(var)
    if len(rep) < 30 and '\n' not in rep:
        return rep
    try:
        return f"<{type(var).__name__}, shape={var.shape}>"
    except AttributeError:
        pass
    try:
        return f"<{type(var).__name__}, len={len(var)}>"
    except TypeError:
        pass
    return f"<{type(var).__name__}>"


def escape_html(text: str) -> str:
    """
    Escape HTML special characters in a string.
    """
    return text.replace('&', '&amp;').replace('<', '&lt;').replace('>', '&gt;'). \
        replace('"', '&quot;').replace("'", '&#39;')


def add_header_and_footer_lines(header: str, text: str, width: int = 30, line: str = '-'):
    """
    Return:
    <-       width      ->

    ------- HEADER -------
    text
    ----------------------
    """
    header_line = f' {header} '.center(width, line)
    footer_line = line * width
    return f'{header_line}\n{text}\n{footer_line}'



from dataclasses import dataclass
from pathlib import Path
from typing import List, Tuple, Optional, Iterable

from data_to_paper.utils.types import ListBasedSet


@dataclass(frozen=True)
class DataframeOperation:
    id: int


@dataclass(frozen=True)
class FileDataframeOperation(DataframeOperation):
    file_path: Optional[str]
    columns: Optional[Iterable]

    @property
    def filename(self):
        if self.file_path is None:
            return None
        return Path(self.file_path).name


@dataclass(frozen=True)
class CreationDataframeOperation(FileDataframeOperation):
    created_by: Optional[str]


@dataclass(frozen=True)
class SaveDataframeOperation(FileDataframeOperation):
    pass


@dataclass(frozen=True)
class SeriesDataframeOperation(DataframeOperation):
    series_name: str


@dataclass(frozen=True)
class AddSeriesDataframeOperation(SeriesDataframeOperation):
    pass


@dataclass(frozen=True)
class RemoveSeriesDataframeOperation(SeriesDataframeOperation):
    pass


@dataclass(frozen=True)
class ChangeSeriesDataframeOperation(SeriesDataframeOperation):
    pass


class DataframeOperations(List[DataframeOperation]):

    def get_read_ids(self) -> ListBasedSet[int]:
        return ListBasedSet(operation.id for operation in self
                            if isinstance(operation, CreationDataframeOperation) and operation.file_path is not None)

    def get_changed_ids(self) -> ListBasedSet[int]:
        return ListBasedSet(operation.id for operation in self if isinstance(operation, SeriesDataframeOperation))

    def get_saved_ids(self) -> ListBasedSet[int]:
        return ListBasedSet(operation.id for operation in self if isinstance(operation, SaveDataframeOperation))

    def get_saved_ids_filenames(self) -> ListBasedSet[Tuple[int, str]]:
        return ListBasedSet((operation.id, operation.filename) for operation in self
                            if isinstance(operation, SaveDataframeOperation))

    def get_read_filename(self, id_: int) -> Optional[str]:
        return next((operation.filename for operation in self
                     if isinstance(operation, CreationDataframeOperation) and operation.id == id_), None)

    def get_read_changed_but_unsaved_ids(self):
        return self.get_read_ids() & self.get_changed_ids() - self.get_saved_ids()

    def get_read_filenames_from_ids(self, ids: Iterable[int]) -> ListBasedSet[Optional[str]]:
        return ListBasedSet(operation.filename for operation in self
                            if operation.id in ids and isinstance(operation, CreationDataframeOperation))

    def get_creation_columns(self, id_: int) -> Optional[List[str]]:
        return next((operation.columns for operation in self
                     if isinstance(operation, CreationDataframeOperation) and operation.id == id_), None)

    def get_save_columns(self, id_: int) -> Optional[List[str]]:
        return next((operation.columns for operation in self
                     if isinstance(operation, SaveDataframeOperation) and operation.id == id_), None)

    def get_changed_columns(self, id_: int) -> List[str]:
        return [operation.series_name for operation in self
                if isinstance(operation, ChangeSeriesDataframeOperation) and operation.id == id_]



from data_to_paper.base_cast.types import Profile, Algorithm


Performer = Profile(
    agent_name='Performer',
    name='Sheldon',
    title='me!',
    description='',
    image_file='',
    algorithm=Algorithm.GPT,
)

Director = Profile(
    agent_name='Director',
    name='USER',
    title='director',
    description='',
    image_file='',
    algorithm=Algorithm.PRE_PROGRAMMED,
)

DataExplorer = Profile(
    agent_name='DataExplorer',
    name='Lenny',
    title='data explorer',
    description='',
    image_file='',
    algorithm=Algorithm.PRE_PROGRAMMED,
)

GoalReviewer = Profile(
    agent_name='GoalReviewer',
    name='Prof. Proton',
    title='goal reviewer',
    description='',
    image_file='',
    algorithm=Algorithm.PRE_PROGRAMMED,
)

DataPreprocessor = Profile(
    agent_name='DataPreprocessor',
    name='Leslie',
    title='data preprocessor',
    description='',
    image_file='',
    algorithm=Algorithm.PRE_PROGRAMMED,
)

PlanReviewer = Profile(
    agent_name='PlanReviewer',
    name='Amy',
    title='research plan reviewer',
    description='',
    image_file='',
    algorithm=Algorithm.GPT,
)

Debugger = Profile(
    agent_name='Debugger',
    name='Raj',
    title='debugger',
    description='',
    image_file='',
    algorithm=Algorithm.PRE_PROGRAMMED,
)

InterpretationReviewer = Profile(
    agent_name='InterpretationReviewer',
    name='Bernie',
    title='interpretation reviewer',
    description='',
    image_file='',
    algorithm=Algorithm.GPT,
)

Writer = Profile(
    agent_name='Writer',
    name='Penny',
    title='scientific writer',
    description='',
    image_file='',
    algorithm=Algorithm.GPT,
)

CitationExpert = Profile(
    agent_name='CitationExpert',
    name='Stewie',
    title='Great with literature citations',
    description='',
    image_file='',
    algorithm=Algorithm.GPT,
)

TableExpert = Profile(
    agent_name='TableExpert',
    name='Howie',
    title='Data-presentation expert',
    description='',
    image_file='',
    algorithm=Algorithm.GPT,
)



from __future__ import annotations

import os
import shutil
import zipfile
import pandas as pd

from dataclasses import dataclass, field
from pathlib import Path
from typing import Optional, List, Union
from unidecode import unidecode

from data_to_paper.code_and_output_files.file_view_params import ViewPurpose
from data_to_paper.env import FOLDER_FOR_RUN
from data_to_paper.latex.clean_latex import wrap_as_latex_code_output
from data_to_paper.utils.file_utils import run_in_directory, clear_directory
from data_to_paper.utils.mutable import Mutable
from data_to_paper.code_and_output_files.referencable_text import NumericReferenceableText, \
    hypertarget_if_referencable_text_product, ReferencableTextProduct

TEXT_EXTS = ['.txt', '.md', '.csv', '.xls', '.xlsx']


@dataclass(frozen=True)
class DataFileDescription:
    file_path: str  # relative to the data directory.  should normally just be the file name
    description: Optional[Union[str, ReferencableTextProduct]] = None  # a user provided description of the file
    originated_from: Optional[str] = None  # None for raw file
    is_binary: Optional[bool] = None  # None for auto based on file extension

    def get_is_binary(self):
        """
        Return True if the file is binary.
        """
        if self.is_binary is not None:
            return self.is_binary

        return Path(self.file_path).suffix not in TEXT_EXTS

    def is_excel(self):
        return Path(self.file_path).suffix in ['.xlsx', '.xls']

    def get_file_header(self, num_lines: int = 4):
        """
        Return the first `num_lines` lines of the file (if they exist).
        """
        if self.is_excel():
            # go over all sheets and return all of them:
            df = pd.read_excel(self.file_path, sheet_name=None)
            s = f'This is an Excel file with {len(df)} sheets. Here is the first few rows for each sheet:\n\n'
            for sheet_name in df.keys():
                s += f'### Sheet: "{sheet_name}"\n'
                s += f'```output\n{df[sheet_name].head(num_lines).to_string(index=False)}\n```\n'
            s += '\n'
            return s

        with open(self.file_path) as f:
            head = []
            for _ in range(num_lines):
                try:
                    head.append(next(f))
                except StopIteration:
                    break
                except UnicodeDecodeError:
                    head.append('UnicodeDecodeError\n')
            return f'Here are the first few lines of the file:\n' \
                   f'```output\n{"".join(head)}\n```\n'

    def pretty_repr(self, num_lines: int = 4, view_purpose: ViewPurpose = None, file_num: Optional[int] = None) -> str:
        if file_num is not None:
            s = f'### File {file_num}: "{self.file_path}"\n'
        else:
            s = f'### "{self.file_path}"\n'
        description = self.description
        if description is not None:
            description = hypertarget_if_referencable_text_product(description, view_purpose, with_header=False)
            s += f'{description}\n\n'
        if num_lines > 0 and not self.get_is_binary():
            s += self.get_file_header(num_lines)
        return s


class DataFileDescriptions(List[DataFileDescription]):
    """
    A list of data file descriptions.
    """

    def __init__(self, *args, data_folder: Optional[Union[str, Path]] = None,
                 general_description: Optional[Union[str, ReferencableTextProduct]] = None,
                 **kwargs):
        super().__init__(*args, **kwargs)
        self.data_folder = data_folder
        self.general_description = general_description

    @classmethod
    def from_other(cls, other: DataFileDescriptions):
        return cls(other, data_folder=other.data_folder, general_description=other.general_description)

    def __str__(self):
        return self.pretty_repr()

    def get_file_description(self, file_path: str) -> DataFileDescription:
        """
        Return the data file description for the given file path.
        """
        for data_file_description in self:
            if data_file_description.file_path == file_path:
                return data_file_description
        raise ValueError(f"Could not find file path {file_path} in data file descriptions.")

    def get_children(self, data_file: DataFileDescription):
        """
        Return the children of the given data file.
        """
        return DataFileDescriptions([data_file_description for data_file_description in self
                                     if data_file_description.originated_from == data_file.file_path],
                                    data_folder=self.data_folder)

    def get_all_raw_files(self):
        """
        Return all the raw files.
        """
        return DataFileDescriptions([data_file_description for data_file_description in self
                                     if data_file_description.originated_from is None],
                                    data_folder=self.data_folder)

    def get_pretty_description_for_file_and_children(
            self, data_file: DataFileDescription, index: Mutable = None, view_purpose: ViewPurpose = None):
        """
        Return a pretty description for the given data file and all its children.
        """
        children = self.get_children(data_file)
        index.val += 1
        s = data_file.pretty_repr(0 if children else 4, view_purpose=view_purpose, file_num=index.val) + '\n'
        for child in children:
            s += self.get_pretty_description_for_file_and_children(child, index)
        return s

    def pretty_repr(self, num_lines: int = 4, view_purpose: ViewPurpose = None) -> str:
        s = ''
        if self.general_description is not None:
            s += '## General Description\n'
            s += hypertarget_if_referencable_text_product(self.general_description, view_purpose,
                                                          with_header=False) + '\n'
        with run_in_directory(self.data_folder):
            s += '## Data Files\n'
            if len(self) == 0:
                s += 'There are no data files'
            elif len(self) == 1:
                s += f"The dataset consists of 1 data file:\n\n"
                s += self[0].pretty_repr(num_lines, view_purpose=view_purpose)
            else:
                s += f"The dataset consists of {len(self)} data files:\n\n"
                index = Mutable(0)
                for parent in self.get_all_raw_files():
                    s += self.get_pretty_description_for_file_and_children(parent, index, view_purpose=view_purpose)
            return s

    def get_data_filenames(self):
        return [data_file_description.file_path for data_file_description in self]

    def to_latex(self,
                 section_name: str = 'Data Description',
                 text: str = 'Here is the data description, as provided by the user:',
                 view_purpose: ViewPurpose = None) -> str:
        s = ''
        s += f"\\section{{{section_name}}}\n{text}\n\n"
        s += wrap_as_latex_code_output(
            self.pretty_repr(num_lines=0, view_purpose=view_purpose))
        return s


@dataclass
class CreateDataFileDescriptions:
    """
    Given a project directory, create a temp folder to run in, and
    the data file descriptions.
    """
    project_directory: Path = None
    data_files_str_paths: List[str] = field(default_factory=list)
    data_files_is_binary: List[Optional[bool]] = field(default_factory=list)

    GENERAL_DESCRIPTION_FILENAME = 'general_description.txt'
    DESCRIPTION_FILENAME_EXT = '.description.txt'

    GENERAL_FILE_DESCRIPTION_PREFIX = 'S'
    FILE_DESCRIPTIONS_PREFIXES = ('T', 'U', 'V', 'W', 'X', 'Y', 'Z')

    temp_folder_to_run_in: Path = FOLDER_FOR_RUN

    def _get_description_file_path(self, data_file_path_str: str):
        data_file_path = self._convert_data_file_path_str_to_path(data_file_path_str)
        return self.project_directory / (data_file_path.name + self.DESCRIPTION_FILENAME_EXT)

    def _read_file_description(self, data_file_path_str: str):
        description_file_path = self._get_description_file_path(data_file_path_str)
        if not description_file_path.exists():
            raise FileNotFoundError(f"Description file for {data_file_path_str} not found.")
        return description_file_path.read_text()

    def _read_general_description(self):
        general_description_file_path = self.project_directory / self.GENERAL_DESCRIPTION_FILENAME
        if not general_description_file_path.exists():
            raise FileNotFoundError(f"General description file not found:\n{general_description_file_path}.")
        return general_description_file_path.read_text()

    def _get_hypertarget_prefix(self, file_num: Optional[int] = None, file_name: str = None):
        """
        Return the hypertarget prefix for the given file number.
        file_num: the file number. None refers to the general description.
        """
        return self.GENERAL_FILE_DESCRIPTION_PREFIX if file_num is None \
            else self.FILE_DESCRIPTIONS_PREFIXES[file_num]

    def _convert_description_to_referenceable_text(self, description: str, file_num: Optional[int],
                                                   file_name: str = None):
        """
        file_num: the file number. None refers to the general description.
        """
        return ReferencableTextProduct(
            referencable_text=NumericReferenceableText(
                text=unidecode(description),
                hypertarget_prefix=self._get_hypertarget_prefix(file_num, file_name)),
            name=file_name,
        )

    def _convert_data_file_path_str_to_path(self, data_file_path: str):
        """
        convert to absolute path.
        if relative path, it is relative to the project directory.
        """
        return Path(data_file_path).absolute() if Path(data_file_path).is_absolute() \
            else self.project_directory / data_file_path

    def _get_file_description_referenceable_text(self, data_file_str_path: str, file_num: Optional[int]):
        description = self._read_file_description(data_file_str_path)
        data_file_path = self._convert_data_file_path_str_to_path(data_file_str_path)
        return self._convert_description_to_referenceable_text(description, file_num, data_file_path.name)

    def _get_general_description_referenceable_text(self):
        return self._convert_description_to_referenceable_text(self._read_general_description(), None,
                                                               'General Description')

    def _copy_files_and_get_list_of_data_file_descriptions(self) -> List[DataFileDescription]:
        data_file_descriptions = []
        clear_directory(self.temp_folder_to_run_in)  # clear data folder
        for j, data_file_str_path in enumerate(self.data_files_str_paths):
            data_file_path = self._convert_data_file_path_str_to_path(data_file_str_path)
            data_file_path_zip = data_file_path.with_name(data_file_path.name + '.zip')
            if os.path.exists(data_file_path):
                # copy file to data folder
                shutil.copyfile(data_file_path, self.temp_folder_to_run_in / data_file_path.name)
            elif os.path.exists(data_file_path_zip):
                # unzip file to data folder
                with zipfile.ZipFile(data_file_path_zip, 'r') as zip_ref:
                    zip_ref.extractall(self.temp_folder_to_run_in)
            else:
                raise FileNotFoundError(f"File {data_file_path.name} or {data_file_path.name}.zip "
                                        f"not found in {data_file_path.parent}")

            data_file_descriptions.append(DataFileDescription(
                file_path=data_file_path.name,  # relative to data dir
                description=self._get_file_description_referenceable_text(data_file_str_path, j),
                is_binary=self.data_files_is_binary[j]
            ))
        return data_file_descriptions

    def create_temp_folder_and_get_file_descriptions(self) -> DataFileDescriptions:
        file_descriptions = self._copy_files_and_get_list_of_data_file_descriptions()
        return DataFileDescriptions(
            file_descriptions,
            data_folder=self.temp_folder_to_run_in,
            general_description=self._get_general_description_referenceable_text()
        )

    def get_raw_str_data_file_descriptions(self) -> DataFileDescriptions:
        return DataFileDescriptions(
            [DataFileDescription(
                file_path=data_file_str_path,  # relative to data dir
                description=self._read_file_description(data_file_str_path),
                is_binary=self.data_files_is_binary[j]
            ) for j, data_file_str_path in enumerate(self.data_files_str_paths)],
            data_folder=None,
            general_description=self._read_general_description(),
        )

    def create_file_descriptions(self, general_description: str, data_file_descriptions: List[str]):
        """
        Create the file descriptions.
        """
        (self.project_directory / self.GENERAL_DESCRIPTION_FILENAME).write_text(general_description)
        for j, data_file_str_path in enumerate(self.data_files_str_paths):
            self._get_description_file_path(data_file_str_path).write_text(data_file_descriptions[j])

    def check_files_exist(self):
        for data_file_str_path in self.data_files_str_paths:
            data_file_path = self._convert_data_file_path_str_to_path(data_file_str_path)
            data_file_path_zip = data_file_path.with_name(data_file_path.name + '.zip')
            if not data_file_path.exists() and not data_file_path_zip.exists():
                raise FileNotFoundError(f"Data file {data_file_path} not found.")



from typing import Any, Optional, Tuple

import numpy as np
from pandas import DataFrame, Series

from data_to_paper.code_and_output_files.referencable_text import label_numeric_value
from data_to_paper.run_gpt_code.overrides.dataframes.df_methods import STR_FLOAT_FORMAT
from data_to_paper.run_gpt_code.overrides.dataframes.utils import df_to_llm_readable_csv, df_to_latex_with_value_format
from data_to_paper.run_gpt_code.overrides.pvalue import is_p_value, PValue
from data_to_paper.utils.numerics import is_lower_eq


def _label_p_value(p_value):
    if is_p_value(p_value):
        s = str(p_value)
        st_sign = PValue.ON_STR.st_sign()
        if st_sign and s.startswith(st_sign):
            return st_sign + label_numeric_value(s[len(st_sign):])
        return label_numeric_value(s)
    return p_value


def _get_formatters(should_format: bool):
    if not should_format:
        return {}
    return dict(
        numeric_formater=lambda x: label_numeric_value(STR_FLOAT_FORMAT(x)),
        object_formatter=_label_p_value
    )


def describe_value(value: Any) -> str:
    """
    Describe the value in a way that can be used as a short string.
    """
    if value is None:
        return 'None'
    if isinstance(value, str):
        return f"'{value}'"
    if isinstance(value, bool):
        return str(value)
    if isinstance(value, (int, float)):
        return str(value)
    if isinstance(value, list):
        return f'[{", ".join(describe_value(v) for v in value)}]'
    if isinstance(value, dict):
        return f'{{{", ".join(f"{describe_value(k)}: {describe_value(v)}" for k, v in value.items())}}}'
    if isinstance(value, tuple):
        return f'({", ".join(describe_value(v) for v in value)})'
    if isinstance(value, DataFrame):
        return f'DataFrame(shape={value.shape}, columns={value.columns})'
    if isinstance(value, Series):
        return f'Series(shape={value.shape}, type={value.dtype})'
    if isinstance(value, np.ndarray):
        return f'np.ndarray(shape={value.shape}, dtype={value.dtype})'
    return str(value)


def describe_df(df: DataFrame, max_rows_and_columns_to_show: Tuple[Optional[int], Optional[int]],
                should_format: bool = True) -> str:
    """
    Describe the DataFrame in a way that can be used as a short string.
    """
    max_rows, max_columns = max_rows_and_columns_to_show
    num_lines = len(df)
    num_columns = len(df.columns)
    if not is_lower_eq(num_columns, max_columns):
        return f'DataFrame(shape={df.shape})'
    if not is_lower_eq(num_lines, max_rows):
        df = df.head(3)
    s = df_to_llm_readable_csv(df, **_get_formatters(should_format))
    if not is_lower_eq(num_lines, max_rows):
        s += f'\n... total {num_lines} rows'
    return s


def df_to_numerically_labeled_latex(df, should_format: bool = True, **kwargs):
    """
    Get latex representation of a DataFrame with numeric values labeled.
    Label the numeric values with @@<...>@@ - to allow converting to ReferenceableText.
    """
    return df_to_latex_with_value_format(df, **_get_formatters(should_format), **kwargs)



from typing import Optional, Dict, Tuple, Union, List, Iterable

import pandas as pd
from matplotlib import pyplot as plt
from pathlib import Path

from data_to_paper.latex.clean_latex import process_latex_text_and_math, replace_special_latex_chars
from data_to_paper.run_gpt_code.overrides.pvalue import PValueToStars, convert_p_values_to_floats, \
    pvalue_on_str_for_latex
from data_to_paper.text.text_formatting import escape_html
from data_to_paper.utils.check_type import raise_on_wrong_func_argument_types_decorator
from data_to_paper.utils.multi_process import run_func_in_separate_process
from data_to_paper.run_gpt_code.config import configure_matplotlib

from .df_plot_with_pvalue import df_plot_with_pvalue, get_description_of_plot_creation
from .matplotlib_utils import get_axis_parameters, AxisParameters, fit_fig_to_axes
from .note_and_legend import convert_note_and_glossary_to_html, convert_note_and_glossary_to_latex_figure_caption
from .utils import convert_to_latex_comment, convert_filename_to_label
from .consts import ALLOWED_PLOT_KINDS, FIG_SIZE_INCHES, AXES_SIZE_INCHES, FIG_DPI


@raise_on_wrong_func_argument_types_decorator
def df_to_figure(df: pd.DataFrame, filename: Optional[str],
                 caption: Optional[str] = None,
                 label: Optional[str] = None,
                 note: Optional[str] = None,
                 glossary: Optional[Dict[str, str]] = None,
                 create_fig: bool = False,
                 is_html: bool = False,
                 figure_folder: Optional[Path] = None,
                 should_format: bool = False,
                 max_rows_and_columns_to_show: Tuple[Optional[int], Optional[int]] = (None, None),
                 **kwargs):
    """
    Create a matplotlib figure embedded in a LaTeX figure with a caption and label.
    """
    label = convert_filename_to_label(filename, label)
    label = 'figure:' + label

    fig_filename = filename + '.png'
    figure_path = figure_folder / fig_filename if figure_folder else fig_filename

    kind = kwargs.get('kind', 'bar')
    if kind not in ALLOWED_PLOT_KINDS:
        raise ValueError(f'`kind` must be one of {ALLOWED_PLOT_KINDS}, but got {repr(kind)}.')

    if create_fig:
        create_fig_for_df_to_figure_and_get_axis_parameters(df, filepath=figure_path, **kwargs)

    index = kwargs.get('use_index', True)
    label = label or ''

    glossary = {} if glossary is None else glossary.copy()
    if kwargs.get('x_p_value') or kwargs.get('y_p_value'):
        glossary['Significance'] = PValueToStars().get_conversion_legend_text()

    caption = caption or ''
    if is_html:
        note_and_glossary = convert_note_and_glossary_to_html(df, note, glossary, index)
        caption_note_and_glossary = escape_html(caption) + '<br>' + note_and_glossary
        description = get_description_of_plot_creation(df, fig_filename, kwargs, is_html=True,
                                                       should_format=should_format,
                                                       max_rows_and_columns_to_show=max_rows_and_columns_to_show)
        s = get_figure_and_caption_as_html(figure_path, caption_note_and_glossary.strip())
        s += description
    else:
        note_and_glossary = convert_note_and_glossary_to_latex_figure_caption(df, note, glossary, index)
        caption_note_and_glossary = replace_special_latex_chars(caption) + '\n' + note_and_glossary
        with pvalue_on_str_for_latex():
            description = get_description_of_plot_creation(df, fig_filename, kwargs, is_html=False,
                                                           should_format=should_format,
                                                           max_rows_and_columns_to_show=max_rows_and_columns_to_show)
        s = get_figure_and_caption_as_latex(fig_filename, caption_note_and_glossary.strip(), label)
        s += '\n' + convert_to_latex_comment(description)
    return s


def _replace_with_str_no_underscores(s):
    if s is None:
        return s
    if not isinstance(s, str):
        if isinstance(s, Iterable):
            return type(s)([_replace_with_str_no_underscores(col) for col in s])
        s = str(s)
    return s.replace('_', ' ')


def replace_df_column_names_to_str_with_no_underscores(df: pd.DataFrame, **kwargs):

    # find all the underscores in the column names:
    df = df.copy()
    df.columns = _replace_with_str_no_underscores(df.columns)

    for arg in ['x', 'y', 'yerr', 'xerr', 'x_ci', 'y_ci', 'x_p_value', 'y_p_value']:
        if arg in kwargs:
            kwargs[arg] = _replace_with_str_no_underscores(kwargs[arg])
    return df, kwargs


def run_create_fig_for_df_to_figure_and_get_axis_parameters(df: pd.DataFrame, filepath: Optional[Path] = None,
                                                            in_separate_process: bool = True, **kwargs) -> List[str]:
    """
    Run the `create_fig_for_df_to_figure` function in a separate process.
    Otherwise we get killing of the kernel due to matplotlib issues.
    """
    return run_func_in_separate_process(create_fig_for_df_to_figure_and_get_axis_parameters, df, filepath,
                                        in_separate_process=in_separate_process, **kwargs)


def create_fig_for_df_to_figure_and_get_axis_parameters(df: pd.DataFrame, filepath: Optional[Path] = None,
                                                        **kwargs) -> AxisParameters:
    configure_matplotlib()

    # The figure size in inches is not consequential, because we use fit_fig_to_axes
    fig = plt.figure(figsize=FIG_SIZE_INCHES, dpi=FIG_DPI.val)

    ax = fig.add_axes((0.1, 0.1, AXES_SIZE_INCHES[0] / FIG_SIZE_INCHES[0], AXES_SIZE_INCHES[1] / FIG_SIZE_INCHES[1]))

    df = convert_p_values_to_floats(df.copy())

    # Replace underscores with spaces in the column names.
    # Otherwise df.plot() may skip the column with underscores in the name.
    df, kwargs = replace_df_column_names_to_str_with_no_underscores(df, **kwargs)

    df_plot_with_pvalue(df, ax=ax, **kwargs)

    # Adjusts subplot parameters to give the plot more room
    fit_fig_to_axes(fig, fit_width=True, fit_height=True, margin_pixels=10)

    if filepath:
        fig.savefig(filepath)

    axis_parameters = get_axis_parameters(ax)

    # figure cleanup
    ax.clear()
    fig.clear()
    plt.close(fig)
    plt.close('all')

    return axis_parameters


def get_figure_and_caption_as_latex(filename: str, caption: str, label: str) -> str:
    """
    Save a figure with a caption and label.
    """
    caption = process_latex_text_and_math(caption)
    latex = f"""
\\begin{{figure}}[htbp]
\\centering
\\includegraphics{{{filename}}}
\\caption{{{caption}}}
\\label{{{label}}}
\\end{{figure}}
"""
    return latex.strip()


def get_figure_and_caption_as_html(filepath: Union[str, Path], caption: str, width: int = 750):
    """
    Save a figure with a caption and label.
    """
    if width is None:
        width_str = ''
    else:
        width_str = f'width="{width}"'
    html = f"""
        <div>
        <img src="{filepath}" alt="{caption}" {width_str} />
        <p>{caption}</p>
        </div>
    """
    return html



from dataclasses import dataclass

from data_to_paper.run_gpt_code.run_issues import RunIssue, CodeProblem


@dataclass
class UnAllowedDataframeMethodCall(RunIssue):
    method_name: str = ''
    issue: str = "Your code uses the dataframe method `{method_name}`, which is not allowed."
    comment: str = 'Code uses forbidden method {method_name}'
    code_problem: CodeProblem = CodeProblem.RuntimeError


def raise_on_call(*args, original_method=None, on_change=None, **kwargs):
    raise UnAllowedDataframeMethodCall.from_current_tb(method_name=original_method.__name__)



import operator

import numpy as np


class OperatorValue:
    OPERATORS_RETURNING_NEW_OBJECT = {}

    UNARY_METHODS_TO_OPERATORS = {
        '__neg__': operator.neg,
        '__pos__': operator.pos,
        '__abs__': operator.abs,
        '__invert__': operator.invert,
        '__round__': round,
        '__float__': float,
        '__floor__': np.floor,
        '__ceil__': np.ceil,
        '__trunc__': np.trunc,
        '__hash__': hash,
        '__format__': format,
        '__str__': str,
        '__repr__': repr,
        '__bool__': bool,
    }

    BINARY_METHODS_TO_OPERATORS = {
        '__lt__': operator.lt,
        '__le__': operator.le,
        '__eq__': operator.eq,
        '__ne__': operator.ne,
        '__gt__': operator.gt,
        '__ge__': operator.ge,
        '__add__': operator.add,
        '__sub__': operator.sub,
        '__mul__': operator.mul,
        '__truediv__': operator.truediv,
        '__floordiv__': operator.floordiv,
        '__mod__': operator.mod,
        '__pow__': operator.pow,
        '__lshift__': operator.lshift,
        '__rshift__': operator.rshift,
        '__and__': operator.and_,
        '__xor__': operator.xor,
        '__or__': operator.or_,
        '__radd__': operator.add,
        '__rsub__': operator.sub,
        '__rmul__': operator.mul,
        '__rtruediv__': operator.truediv,
        '__rfloordiv__': operator.floordiv,
        '__rmod__': operator.mod,
        '__rpow__': operator.pow,
        '__rlshift__': operator.lshift,
        '__rrshift__': operator.rshift,
        '__rand__': operator.and_,
        '__rxor__': operator.xor,
        '__ror__': operator.or_,
    }

    def __init__(self, value):
        self.value = value

    def _get_new_object(self, value):
        return self.__class__(value)

    def _apply_post_operator(self, op, method_name, value):
        if method_name in self.OPERATORS_RETURNING_NEW_OBJECT:
            return self._get_new_object(value)
        return value

    def _raise_on_forbidden_operator(self, op, method_name):
        if op is None:
            raise NotImplementedError(f'Operator {method_name} is not allowed on {self.__class__.__name__}')

    def _binary_op(self, other, op, method_name):
        self._raise_on_forbidden_operator(op, method_name)
        if isinstance(other, OperatorValue):
            other = other.value
        return self._apply_post_operator(op, method_name, op(self.value, other))

    def _unary_op(self, op, *args, method_name=None, **kwargs):
        self._raise_on_forbidden_operator(op, method_name)
        return self._apply_post_operator(op, method_name, op(self.value, *args, **kwargs))

    @classmethod
    def add_methods(cls):
        for method_name, op in cls.BINARY_METHODS_TO_OPERATORS.items():
            def method(self, other, op=op, method_name=method_name):
                return self._binary_op(other, op, method_name)
            setattr(cls, method_name, method)

        for method_name, op in cls.UNARY_METHODS_TO_OPERATORS.items():
            def method(self, *args, op=op, method_name=method_name, **kwargs):
                return self._unary_op(op, *args, method_name=method_name, **kwargs)
            setattr(cls, method_name, method)
        return cls


OperatorValue.add_methods()



class ScholarServer:
    def __init__(self, server: str):
        self.server = server

    def get_server_instance(self):
        """
        Get the server instance.
        """
        if self.server == "Crossref":
            from data_to_paper.servers.crossref import CROSSREF_SERVER_CALLER

            return CROSSREF_SERVER_CALLER
        else:
            from data_to_paper.servers.semantic_scholar import (
                SEMANTIC_SCHOLAR_SERVER_CALLER,
            )

            return SEMANTIC_SCHOLAR_SERVER_CALLER



from dataclasses import dataclass
from typing import Optional

from data_to_paper.text.formatted_sections import FormattedSections


@dataclass
class FailedExtractingBlock(Exception):
    content_name: str  # e.g. "code", "latex"
    requested_label: Optional[str]  # e.g. "python", "latex"

    def block_name(self):
        if self.requested_label is None:
            return f"triple-backtick block"
        else:
            return f'triple-backtick "{self.requested_label}" block'

    def get_problem(self):
        return ''

    def __str__(self):
        return f"Error extracting '{self.block_name()}' block:\n" + self.get_problem()


@dataclass
class NoBlocksFailedExtractingBlock(FailedExtractingBlock):
    def get_problem(self):
        return f"You did not send any triple-backtick block.\n" \
               f"Please try again, making sure the {self.content_name} is enclosed within {self.block_name()}."


@dataclass
class MultiBlocksFailedExtractingBlock(FailedExtractingBlock):
    num_blocks: int

    def get_problem(self):
        return f"You sent {self.num_blocks} triple-backtick blocks. " \
               f"Please send the {self.content_name} as a single {self.block_name()}."


@dataclass
class IncompleteBlockFailedExtractingBlock(FailedExtractingBlock):
    def get_problem(self):
        return f"You sent an incomplete triple-backtick block. Please try again."


@dataclass
class WrongLabelFailedExtractingBlock(FailedExtractingBlock):
    label: str

    def get_problem(self):
        return f'Your sent a "{self.label}" block. ' \
               f'Please send your {self.content_name} as a "{self.block_name()}".'


def add_label_to_first_triple_quotes_if_missing(content: str, requested_label: str) -> str:
    """
    Add "python" to triple-backtick if missing.
    """
    formatted_sections = FormattedSections.from_text(content)
    first_block = formatted_sections.get_first_block()
    if first_block is not None:
        if first_block.label.lower() == '':
            first_block.label = requested_label
    return formatted_sections.to_text()


def extract_content_of_triple_quote_block(text: str, content_name: str, requested_label: Optional[str] = None) -> str:
    formatted_sections = FormattedSections.from_text(text)
    block_sections = formatted_sections.get_all_blocks()
    if len(block_sections) == 0:
        raise NoBlocksFailedExtractingBlock(content_name, requested_label)
    if len(block_sections) > 1:
        raise MultiBlocksFailedExtractingBlock(content_name, requested_label, len(block_sections))
    block_section = block_sections[0]
    if not block_section.is_complete:
        raise IncompleteBlockFailedExtractingBlock(content_name, requested_label)
    if requested_label is not None and block_section.label and block_section.label.lower() != requested_label.lower():
        raise WrongLabelFailedExtractingBlock(content_name, requested_label, block_section.label)
    return block_section.section


def extract_code_from_text(text: str) -> str:
    """
    Extract Python code from text.
    Raise FailedExtractingBlock if the code cannot be extracted.
    """
    return extract_content_of_triple_quote_block(text, content_name='code', requested_label="python")



from types import ModuleType
from typing import Callable, Union, Tuple, Any, Optional

from .base_run_contexts import MultiRunContext
from .exceptions import FailedRunningCode

from data_to_paper.utils.types import ListBasedSet
from data_to_paper.base_steps import BaseCodeProductsGPT


def run_code_in_context(code: Union[str, Callable, ModuleType], code_converser: BaseCodeProductsGPT,
                        run_folder: str = None
                        ) -> Tuple[Any, ListBasedSet[str], MultiRunContext, Optional[FailedRunningCode]]:
    """
    This function allows to manually run code created by the LLM model.
    We provide the code to run and the code_converser object that was used to create the code.
    The code is run in the correct context and the result is returned.

    The `code` can be provided as:
    - a string of code
        To run the code, we save it to a .py file and use the importlib to import it.
    - a function
        The function is called in the current context.
    - a module
        The module is reloaded and the function is called in the current context.

    run_folder:
        The folder where the code will be run (where the data is stored).
        If not provided, the code will run in the default run_folder (FOLDER_FOR_RUN).
    """
    debugger = code_converser.get_debugger(
        data_filenames='all',  # do not limit the data files
        data_folder=run_folder,
        is_new_conversation=None,
    )
    code_runner = debugger.get_code_runner()
    if run_folder is not None:
        code_runner.run_folder = run_folder
    return code_runner.run(code)



"""
Run this script to check if the API keys for the LLMServer and Semantic Scholar are valid.
"""
from data_to_paper.latex.check_dependencies import check_all_pdflatex_dependencies
from data_to_paper.latex.latex_to_html import check_pandoc_is_installed
from data_to_paper.scripts.run import get_version
from data_to_paper.servers.check_connection import check_all_servers


def check_resources():
    print(f'\n\n** data-to-paper version {get_version()} **\n')
    print('RESOURCE CHECKS\n')
    print('Checking external server connections, API keys, and installation dependencies...\n')
    check_all_servers()
    check_all_pdflatex_dependencies()
    check_pandoc_is_installed()



import pickle
from dataclasses import dataclass, field
from pathlib import Path
from types import ModuleType

import os
import importlib

from typing import Optional, Type, Tuple, Any, Union, Iterable, Dict, Callable

from data_to_paper.env import DEBUG_MODE
from data_to_paper.utils.types import ListBasedSet
from data_to_paper.code_and_output_files.output_file_requirements import OutputFileRequirements

from .base_run_contexts import MultiRunContext
from .attr_replacers import PreventCalling
from .config import configure_matplotlib
from .run_contexts import PreventFileOpen, ModifyImport, WarningHandler, IssueCollector, \
    TrackCreatedFiles, RunInDirectory

from .exceptions import FailedRunningCode, BaseRunContextException
from .user_script_name import MODULE_NAME, module_filename
from .run_issues import RunIssue

from data_to_paper import llm_created_scripts
module_dir = os.path.dirname(llm_created_scripts.__file__)
module_default_filepath = os.path.join(module_dir, module_filename)

USING_MATPLOTLIB_IN_GPT_CODE = False


def save_code_to_module_file(code: str = None):
    code = code or '# empty module\n'
    with open(module_default_filepath, "w", encoding='utf-8') as f:
        f.write(code)


def generate_empty_code_module_object() -> ModuleType:
    """
    Generate module object with the given code and return it.
    """
    save_code_to_module_file()
    return importlib.import_module(llm_created_scripts.__name__ + '.' + MODULE_NAME)


def is_serializable(x):
    """
    Check if x is serializable so that it can be transferred between processes.
    """
    try:
        pickle.dumps(x)
        return True
    except Exception:
        return False


@dataclass
class CodeRunner:
    """
    Run the provided code and report exceptions or specific warnings.
    """
    warnings_to_ignore: Optional[Iterable[Type[Warning]]] = \
        (DeprecationWarning, ResourceWarning, PendingDeprecationWarning, FutureWarning)
    warnings_to_raise: Optional[Iterable[Type[Warning]]] = ()
    warnings_to_issue: Optional[Iterable[Type[Warning]]] = ...  # `...` for all that are not ignored or raised

    forbidden_modules_and_functions: Iterable[Tuple[Any, str, bool]] = \
        (
            # Module, function, create RunIssue (True) or raise exception (False)
            ('builtins', 'print', True),
            ('builtins', 'input', False),
            ('builtins', 'eval', False),
            ('builtins', 'exit', False),
            ('builtins', 'quit', False),
            ('matplotlib.pyplot', 'show', True),  # 'matplotlib.pyplot' is a string because it is not always installed
    )
    modified_imports: Tuple[Tuple[str, Optional[str]]] = (
            ('os', None),
            ('sys', None),
            ('subprocess', None),
            ('shutil', None),
    )

    # File lists can include wildcards, e.g. '*.py' or '**/*.py'. () means no files. None means all files.

    # Allowed in `open` for reading/writing
    allowed_open_read_files: Optional[Iterable[str]] = ()  # 'all' means all files. () means no files.
    allowed_open_write_files: Optional[Iterable[str]] = None  # 'all': all. `None`: Based on output_file_requirements

    # Allowed new files. Assessed at end of run. If None then all files are allowed.
    output_file_requirements: Optional[OutputFileRequirements] = OutputFileRequirements()

    run_folder: Union[Path, str] = field(default_factory=Path)

    additional_contexts: Optional[Dict[str, Any]] = field(default_factory=dict)

    _multi_context: MultiRunContext = None

    _module: Optional[ModuleType] = None

    def _get_or_create_multi_context(self) -> MultiRunContext:
        if self._multi_context is not None:
            return self._multi_context

        allowed_open_write_files = self.allowed_open_write_files if self.allowed_open_write_files is not None \
            else self.output_file_requirements.get_all_allowed_created_filenames()

        # Mandatory contexts:
        contexts = {
            'RunInDirectory': RunInDirectory(folder=self.run_folder),
            'IssueCollector': IssueCollector(),
            'TrackCreatedFiles': TrackCreatedFiles(output_file_requirements=self.output_file_requirements),
        }

        # Optional builtin contexts:
        if self.forbidden_modules_and_functions is not None:
            contexts['PreventCalling'] = PreventCalling(modules_and_functions=self.forbidden_modules_and_functions)
        if self.modified_imports is not None:
            contexts['ModifyImport'] = ModifyImport(modified_imports=self.modified_imports)
        if not (self.allowed_open_read_files is None and allowed_open_write_files is None):
            contexts['PreventFileOpen'] = PreventFileOpen(allowed_read_files=self.allowed_open_read_files,
                                                          allowed_write_files=allowed_open_write_files,
                                                          allowed_write_folder=self.run_folder)
        if not (self.warnings_to_raise is None and self.warnings_to_issue is None and self.warnings_to_ignore is None):
            contexts['WarningHandler'] = WarningHandler(categories_to_raise=self.warnings_to_raise,
                                                        categories_to_issue=self.warnings_to_issue,
                                                        categories_to_ignore=self.warnings_to_ignore)
        # Additional custom contexts:
        if self.additional_contexts is not None:
            for context_name, context in self.additional_contexts.items():
                assert context_name not in contexts, f"Context name {context_name} already exists."
                contexts[context_name] = context

        # name all contexts
        for name, context in contexts.items():
            context.name = name

        self._multi_context = MultiRunContext(contexts=contexts)
        return self._multi_context

    def run(self, code: Union[str, Callable, ModuleType]
            ) -> Tuple[Any, ListBasedSet[str], MultiRunContext, Optional[FailedRunningCode]]:
        """
        Run the provided code and report exceptions or specific warnings.

        `code` can be provided as:
        - a string of code to run,
            To run the code, we save it to a .py file and use the importlib to import it.
        - a function to run,
            The function is called in the current context.
        - a module to run,
            The module is reloaded and the function is called in the current context.

        Returns:
            result: the result of a call to a function in the code, None if no function was called.
            created_files: the files that were created during the run.
            multi_context: the multi context that was used during the run.
            exception: an exception that was raised during the run, None if no exception was raised.
        """
        if USING_MATPLOTLIB_IN_GPT_CODE:
            configure_matplotlib()

        if isinstance(code, str):
            self._module = generate_empty_code_module_object()
            save_code_to_module_file(code)
        elif isinstance(code, ModuleType):
            self._module = code
        else:
            self._module = None

        multi_context = self._get_or_create_multi_context()
        exception = None
        result = None
        try:
            with multi_context:
                try:
                    if self._module is None:
                        result = code()
                    else:
                        module = importlib.reload(self._module)
                        result = self._run_function_in_module(module)
                except RunIssue as e:
                    exception = e
                except Exception as e:
                    exception = FailedRunningCode.from_exception(e)

        except BaseRunContextException as e:
            exception = FailedRunningCode.from_exception(e)
        except Exception:
            raise
        finally:
            if not DEBUG_MODE:
                save_code_to_module_file()  # leave the module empty

        for context in multi_context.get_contexts():
            assert is_serializable(context), f"Context {context} is not serializable."

        created_files = multi_context.contexts['TrackCreatedFiles'].created_files
        return result, created_files, multi_context, exception

    def _run_function_in_module(self, module: ModuleType):
        pass



from dataclasses import dataclass
from typing import Iterable, Any, Type, Tuple, Optional, Dict, Collection, List

from data_to_paper.base_steps import DebuggerConverser
from data_to_paper.base_steps.request_code import CodeReviewPrompt
from data_to_paper.code_and_output_files.code_and_output import CodeAndOutput
from data_to_paper.code_and_output_files.file_view_params import ViewPurpose
from data_to_paper.code_and_output_files.output_file_requirements import OutputFileRequirements
from data_to_paper.code_and_output_files.ref_numeric_values import HypertargetFormat, HypertargetPosition, \
    ReferencedValue
from data_to_paper.code_and_output_files.referencable_text import convert_str_to_latex_label
from data_to_paper.latex.latex_doc import LatexDocument
from data_to_paper.research_types.hypothesis_testing.cast import ScientificAgent
from data_to_paper.research_types.hypothesis_testing.coding.base_code_conversers import BaseTableCodeProductsGPT
from data_to_paper.research_types.hypothesis_testing.coding.utils import create_pandas_and_stats_contexts
from data_to_paper.research_types.hypothesis_testing.scientific_products import HypertargetPrefix, ScientificProducts
from data_to_paper.run_gpt_code.attr_replacers import PreventAssignmentToAttrs, PreventCalling, AttrReplacer
from data_to_paper.run_gpt_code.code_runner import CodeRunner
from data_to_paper.run_gpt_code.overrides.dataframes.df_with_attrs import InfoDataFrameWithSaveObjFuncCall
from data_to_paper.run_gpt_code.overrides.pvalue import PValue, OnStr, OnStrPValue
from data_to_paper.run_gpt_code.run_issues import RunIssue, CodeProblem
from data_to_paper.text import dedent_triple_quote_str

from ..analysis.coding import BaseDataFramePickleContentOutputFileRequirement, DataAnalysisDebuggerConverser
from ...check_df_to_funcs.df_checker import check_displayitem_df


@dataclass
class DataframePreventAssignmentToAttrs(PreventAssignmentToAttrs):
    obj_import_str: str = 'pandas.DataFrame'
    forbidden_set_attrs: Iterable[str] = ('columns', 'index')

    def _raise_exception(self, attr, value):
        raise RunIssue.from_current_tb(
            category='Coding: good practices',
            issue=f"To avoid mistakes, please do not directly assign to '{attr}'.",
            code_problem=CodeProblem.NonBreakingRuntimeIssue,
            instructions=f'Use instead `df.rename({attr}=<mapping>, inplace=True)`',
        )


@dataclass(frozen=True)
class TexDisplayitemContentOutputFileRequirement(BaseDataFramePickleContentOutputFileRequirement):
    hypertarget_prefixes: Optional[Tuple[str]] = HypertargetPrefix.LATEX_TABLES.value
    latex_document: Optional[LatexDocument] = None

    def _check_df(self, content: InfoDataFrameWithSaveObjFuncCall) -> List[RunIssue]:
        return check_displayitem_df(content, output_folder=self.output_folder, latex_document=self.latex_document)

    def _convert_view_purpose_to_pvalue_on_str(self, view_purpose: ViewPurpose) -> OnStr:
        return OnStr.SMALLER_THAN

    def _get_hyper_target_format(self, content: InfoDataFrameWithSaveObjFuncCall, filename: str = None,
                                 num_file: int = 0, view_purpose: ViewPurpose = None) -> HypertargetFormat:
        if self._is_figure(content):
            if view_purpose == ViewPurpose.FINAL_INLINE:
                return HypertargetFormat()
            if view_purpose == ViewPurpose.FINAL_APPENDIX:
                return HypertargetFormat(position=HypertargetPosition.ADJACENT, raised=True, escaped=True)
        else:
            if view_purpose == ViewPurpose.FINAL_INLINE:
                return HypertargetFormat(position=HypertargetPosition.ADJACENT, raised=True, escaped=False)
            if view_purpose == ViewPurpose.FINAL_APPENDIX:
                return HypertargetFormat()

        return super()._get_hyper_target_format(content, filename, num_file, view_purpose)

    def _get_block_label(self, filename: str, num_file: int, view_purpose: ViewPurpose) -> str:
        return 'latex'

    def _convert_content_to_labeled_text(self, content: InfoDataFrameWithSaveObjFuncCall, filename: str = None,
                                         num_file: int = 0, view_purpose: ViewPurpose = None) -> str:
        func, args, kwargs = content.get_func_call()
        pvalue_on_str = self._convert_view_purpose_to_pvalue_on_str(view_purpose)
        if view_purpose == ViewPurpose.FINAL_INLINE:
            caption = kwargs.get('caption', '')
            caption_lines = caption.split('\n')
            first_line = caption_lines[0]
            filename = content.get_prior_filename()
            first_line = r"\protect" + ReferencedValue(
                value=first_line,
                label=convert_str_to_latex_label(filename + '.pkl', prefix='file'),
                is_target=False).to_str(HypertargetFormat(position=HypertargetPosition.WRAP))
            caption_lines[0] = first_line
            kwargs = kwargs.copy()
            kwargs['caption'] = '\n'.join(caption_lines)

        with OnStrPValue(pvalue_on_str):
            return func(*args, **kwargs, should_format=True)

    def _get_content_and_header_for_final_inline(
            self, content: InfoDataFrameWithSaveObjFuncCall, filename: str = None, num_file: int = 0, level: int = 3,
            view_purpose: ViewPurpose = ViewPurpose.FINAL_INLINE):
        text, header_refs = self.get_formatted_text_and_header_references(content, filename, num_file, view_purpose)
        return text, f'% {filename}\n' + '\n'.join(header_ref.to_str() for header_ref in header_refs)


@dataclass
class UtilsCodeRunner(CodeRunner):
    modified_imports: Tuple[Tuple[str, Optional[str]]] = CodeRunner.modified_imports + (
        ('my_utils', 'data_to_paper.research_types.hypothesis_testing.coding.displayitems.my_utils'),
    )


@dataclass
class DisplayitemsDebuggerConverser(DataAnalysisDebuggerConverser):
    products: ScientificProducts = None

    def _get_issues_for_missing_dfs(self, code_and_output: CodeAndOutput) -> List[RunIssue]:
        created_pkl_df_files = self.products.get_created_dfs()
        required_tex_files = [file.replace('.pkl', '.tex') for file in created_pkl_df_files]
        created_tex_files = code_and_output.created_files.get_created_content_files()
        missing_tex_files = set(required_tex_files) - set(created_tex_files)
        if len(missing_tex_files):
            return [RunIssue(
                category='Missing output files',
                issue=f"You did not create a tex file for the following tables: {missing_tex_files}",
                instructions=f"Please create a tex file for each table.",
                code_problem=CodeProblem.OutputFileCallingSyntax,
            )]
        return []

    def _get_issues_for_created_output_files(self, code_and_output: CodeAndOutput, contexts) -> List[RunIssue]:
        issues = super()._get_issues_for_created_output_files(code_and_output, contexts)
        if issues:
            return issues
        return self._get_issues_for_missing_dfs(code_and_output)

    def _get_issues_for_df_comments(self, code_and_output: CodeAndOutput, contexts) -> List[RunIssue]:
        return []  # TODO: Implement this method


@dataclass
class CreateDisplayitemsCodeProductsGPT(BaseTableCodeProductsGPT):
    COPY_ATTRIBUTES = ('latex_document', )
    latex_document: LatexDocument = None
    code_step: str = 'data_to_latex'
    debugger_cls: Type[DebuggerConverser] = DisplayitemsDebuggerConverser
    code_runner_cls: Type[CodeRunner] = UtilsCodeRunner
    headers_required_in_code: Tuple[str, ...] = (
        '# IMPORT',
        '# PREPARATION FOR ALL TABLES AND FIGURES',
    )
    phrases_required_in_code: Tuple[str, ...] = \
        ('\nfrom my_utils import df_to_latex, df_to_figure, is_str_in_df, split_mapping, AbbrToNameDef', )

    max_debug_iterations_per_attempt: int = 20
    max_code_revisions: int = 1
    user_agent: ScientificAgent = ScientificAgent.InterpretationReviewer
    background_product_fields: Tuple[str, ...] = \
        ('data_file_descriptions', 'research_goal', 'codes:data_preprocessing', 'codes:data_analysis',
         'created_files_content:data_analysis:df_*.pkl')
    allow_data_files_from_sections: Tuple[Optional[str]] = ('data_analysis', )
    supported_packages: Tuple[str, ...] = ('pandas', 'numpy', 'my_utils')

    def _create_output_file_requirements(self) -> OutputFileRequirements:
        return OutputFileRequirements([
            TexDisplayitemContentOutputFileRequirement(
                'df_*_formatted.pkl', minimal_count=1,
                output_folder=self.output_directory,
                latex_document=self.latex_document)
        ])

    provided_code: str = dedent_triple_quote_str('''
        {df_to_latex_doc}

        {df_to_figure_doc}

        def is_str_in_df(df: pd.DataFrame, s: str):
            return any(s in level for level in getattr(df.index, 'levels', [df.index]) + \t
        getattr(df.columns, 'levels', [df.columns]))

        AbbrToNameDef = Dict[Any, Tuple[Optional[str], Optional[str]]]

        def split_mapping(abbrs_to_names_and_definitions: AbbrToNameDef):
            abbrs_to_names = {abbr: name for abbr, (name, definition) in \t
        abbrs_to_names_and_definitions.items() if name is not None}
            names_to_definitions = {name or abbr: definition for abbr, (name, definition) in \t
        abbrs_to_names_and_definitions.items() if definition is not None}
            return abbrs_to_names, names_to_definitions
        ''')

    mission_prompt: str = dedent_triple_quote_str('''
        Please write a Python code to convert and re-style the "df_*.pkl" dataframes created \t
        by our "{codes:data_analysis}" into nicer latex tables/figures suitable for our scientific paper.

        Your code should use the following 4 custom functions provided for import from `my_utils`: 

        ```python
        {provided_code}
        ```

        Your code should define `all_mapping: AbbrToNameDef`, which maps any original \t
        column and row names that are abbreviated or not self-explanatory to an optional new name, \t
        and an optional definition.

        Then, for each df_*.pkl file, your code should:

        * Rename column and row names: You should provide a new name to any column or row label that is abbreviated \t
        or technical, or that is otherwise not self-explanatory.

        * Provide glossary definitions: You should provide a full definition for any original name \t
        (or modified new name) in the df that satisfies any of the following: 
        - Remains abbreviated, or not self-explanatory, even after renaming.
        - Is an ordinal/categorical variable that requires clarification of the meaning of each of its possible values.
        - Contains unclear notation, like '*' or ':', '_'.
        - Represents a numeric variable that has units, that need to be specified.        

        To avoid re-naming mistakes, you should define for each df a dictionary, \t
        `mapping: AbbrToNameDef`, derived from `all_mapping`, that contains only the labels that are in the df \t
        (use the `is_str_in_df` function). 

        Overall, the code must have the following structure (### are my instructions, not part of the code):

        ```python
        # IMPORT
        import pandas as pd
        from my_utils import df_to_latex, df_to_figure, is_str_in_df, split_mapping, AbbrToNameDef

        # PREPARATION FOR ALL TABLES AND FIGURES
        ### Define mapping for all df labels that need to be renamed and/or glossary defined. For example:
        all_mapping: AbbrToNameDef = {
            # Rename and provide glossary definitions for any abbreviated or not self-explanatory labels:
            'AvgAge': ('Avg. Age', 'Average age, years'),
            'BT': ('Body Temperature', '1: Normal, 2: High, 3: Very High'),

            # Explain and add units:
            'Weight': ('Weight', 'Participant weight, kg'),

            # Provide a full definition for any label that is not self-explanatory:
            'MRSA': ('MRSA', 'Infected with Methicillin-resistant Staphylococcus aureus, 1: Yes, 0: No'),

            # If the table is too wide, provide a shorter label:
            'Too Long Label': ('Short Label', 'Definition of the short label'),

            # Etc.
            ...: (..., ...),
        }
        ### These are of course just examples. 
        ### Consult with the "{data_file_descriptions}" and the "{codes:data_analysis}" 
        ### for choosing the actual labels and their proper scientific names and definitions.

        ## Process df_tag:  
        ### `tag` is a placeholder for the actual name of the df created by the {codes:data_analysis}
        df_tag = pd.read_pickle('df_tag.pkl')

        # Format values:
        ### If not needed, write '# Not Applicable' under the '# Format values:' header.
        ### Rename technical values to scientifically-suitable values. For example:
        df_tag['MRSA'] = df_tag['MRSA'].apply(lambda x: 'Yes' if x == 1 else 'No')

        # Rename rows and columns:
        ### Rename any abbreviated or not self-explanatory df labels to scientifically-suitable names.
        ### Get from the `all_mapping` the labels that are in the focal df:
        mapping = dict((k, v) for k, v in all_mapping.items() if is_str_in_df(df_mrsa_age, k))
        ### If needed, add/change df-specific `mapping` \t
        (typically, `mapping` should be a subset of `all_mapping`, so this step might not be needed). 
        mapping |= {
            'PV': ('P-value', None),
            'CI': ('CI', '95% Confidence Interval'),
            'Sex_Age': ('Age * Sex', 'Interaction term between Age and Sex'),
        }
        abbrs_to_names, glossary = split_mapping(mapping)
        df_tag.rename(columns=abbrs_to_names, index=abbrs_to_names, inplace=True)

        ### Choose whether it is more appropriate to present the data as a table or a figure: 

        ### As a table:
        df_to_latex(
            df_tag, 'df_tag_formatted'
            caption="<choose a caption suitable for a table in a scientific paper>", 
            note="<If needed, add a note to provide any additional information that is not captured in the caption>",
            glossary=glossary)

        ### As a figure:
        df_to_figure(
            df_tag, 'df_tag_formatted',
            caption="<one line heading of the figure (this will get bolded in the scientific papers).>", 
            note="<If needed, add a note with additional information that will appear below the caption.
                  Do not repeat the caption. Do not repeat the glossary. 
                  Do not specify '** < 0.001' (will add automatically)>",
            glossary=glossary,
            kind='bar',
            y='Coefficient',
            y_ci='CI',  # a column with (lower, upper) tuple values.
            y_p_value='PV',  # a column with p-values.
        )


        ## Process df_next_tag:
        ### etc, for each 'df_*.pkl'
        ```

        Avoid the following:
        Do not provide a sketch or pseudocode; write a complete runnable code including all '# HEADERS' sections.
        Do not explicitly create any graphics or tables; use the provided functions.
        Do not send any presumed output examples.
        ''')

    code_review_prompts: Collection[CodeReviewPrompt] = ()

    def __post_init__(self):
        super().__post_init__()
        self.headers_required_in_code += tuple(f'## {file_name.split(".")[0]}'
                                               for file_name in self.products.get_created_dfs())

    def _get_additional_contexts(self) -> Optional[Dict[str, Any]]:
        return create_pandas_and_stats_contexts(allow_dataframes_to_change_existing_series=True,
                                                enforce_saving_altered_dataframes=False) | {
            'CustomPreventMethods': PreventCalling(
                modules_and_functions=(
                    ('pandas', 'to_numeric', False),
                )
            ),
            'CustomPreventAssignmentToAtt': DataframePreventAssignmentToAttrs(
                forbidden_set_attrs=['columns', 'index'],
            ),
            'PValueMessage': AttrReplacer(
                obj_import_str=PValue, attr='error_message_on_forbidden_func',
                wrapper="Calling `{func_name}` on a PValue object is forbidden."
            ),
        }



import re

from dataclasses import dataclass
from functools import partial
from typing import Optional, Dict, Union, Iterable, Collection, Tuple

from pathlib import Path

from data_to_paper.terminate.exceptions import MissingInstallationError
from data_to_paper.latex.clean_latex import process_latex_text_and_math
from data_to_paper.latex.latex_to_pdf import evaluate_latex_num_command, is_pdflatex_package_installed, \
    is_pdflatex_installed, save_latex_and_compile_to_pdf, PDFLATEX_INSTALLATION_INSTRUCTIONS

from data_to_paper.servers.custom_types import Citation
from data_to_paper.text import dedent_triple_quote_str

DEFAULT_PACKAGES = (
    '[utf8]{inputenc}',
    '{hyperref}',
    '{amsmath}',
    '{booktabs}',
    '{multirow}',
    '{threeparttable}',
    '{fancyvrb}',
    '{color}',
    '{listings}',
    '{sectsty}',
    '{graphicx}',
)

DEFAULT_INITIATION_COMMANDS = (r"""
% Default fixed font does not support bold face
\DeclareFixedFont{\ttb}{T1}{txtt}{bx}{n}{12} % for bold
\DeclareFixedFont{\ttm}{T1}{txtt}{m}{n}{12}  % for normal

% Custom colors
\usepackage{color}
\definecolor{deepblue}{rgb}{0,0,0.5}
\definecolor{deepred}{rgb}{0.6,0,0}
\definecolor{deepgreen}{rgb}{0,0.5,0}
\definecolor{cyan}{rgb}{0.0,0.6,0.6}
\definecolor{gray}{rgb}{0.5,0.5,0.5}

% Python style for highlighting
\newcommand\pythonstyle{\lstset{
language=Python,
basicstyle=\ttfamily\footnotesize,
morekeywords={self, import, as, from, if, for, while},              % Add keywords here
keywordstyle=\color{deepblue},
stringstyle=\color{deepred},
commentstyle=\color{cyan},
breaklines=true,
escapeinside={(*@}{@*)},            % Define escape delimiters
postbreak=\mbox{\textcolor{deepgreen}{$\hookrightarrow$}\space},
showstringspaces=false
}}


% Python environment
\lstnewenvironment{python}[1][]
{
\pythonstyle
\lstset{#1}
}
{}

% Python for external files
\newcommand\pythonexternal[2][]{{
\pythonstyle
\lstinputlisting[#1]{#2}}}

% Python for inline
\newcommand\pythoninline[1]{{\pythonstyle\lstinline!#1!}}


% Code output style for highlighting
\newcommand\outputstyle{\lstset{
    language=,
    basicstyle=\ttfamily\footnotesize\color{gray},
    breaklines=true,
    showstringspaces=false,
    escapeinside={(*@}{@*)},            % Define escape delimiters
}}

% Code output environment
\lstnewenvironment{codeoutput}[1][]
{
    \outputstyle
    \lstset{#1}
}
{}

""",
)  # noqa

START_APPENDIX = r"""
\clearpage
\appendix
"""

CITATION_TEMPLATE = r"""
\bibliographystyle{unsrt}
\bibliography{citations}
"""


def replace_scientific_exponent_with_latex(text, with_dollar_signs=True):
    def replace(match):
        base, exponent = match.groups()
        exponent = int(exponent)  # to remove leading zeros
        if base.startswith('-') or base.startswith('+'):
            sign = base[0]
            base = base[1:]
        else:
            sign = ''
        if base == '1':
            s = r'{}10^{{{}}}'.format(sign, exponent)
        else:
            s = r'{}{}\ 10^{{{}}}'.format(sign, base, exponent)
        if with_dollar_signs:
            s = '$' + s + '$'
        return s

    return re.sub(pattern=r'([+-]?[\d.]+)e([+-]?\d+)', repl=replace, string=text)


def get_tabular_block(latex_table: str) -> str:
    """
    Extract the tabular block of the table.
    """
    return re.search(pattern=r'\\begin{tabular}.*\n(.*)\\end{tabular}', string=latex_table, flags=re.DOTALL).group(0)


IS_PDFLATEX_INSTALLED: Optional[bool] = None
MISSING_PDFLATEX_PACKAGES: Optional[bool] = None


@dataclass(frozen=True)
class LatexDocument:
    """
    A class for creating latex documents and compiling them to pdf.
    """

    kind: str = 'article'
    fontsize: int = 11

    section_heading_fontsize: str = 'Large'
    subsection_heading_fontsize: str = 'normalsize'
    subsubsection_heading_fontsize: str = 'normalsize'
    initiation_commands: Tuple[str] = DEFAULT_INITIATION_COMMANDS

    section_numbering: bool = False
    subsection_numbering: bool = False
    subsubsection_numbering: bool = False

    replace_scientific_exponents: bool = True

    author: str = 'data-to-paper'
    packages: Tuple[str] = DEFAULT_PACKAGES

    allow_displayitem_tilde: bool = False

    def _style_section(self, section: str) -> str:
        if not self.section_numbering:
            section = section.replace(r'\section{', r'\section*{')
        else:
            section = section.replace(r'\section*{', r'\section{')

        if not self.subsection_numbering:
            section = section.replace(r'\subsection{', r'\subsection*{')
        else:
            section = section.replace(r'\subsection*{', r'\subsection{')

        if not self.subsubsection_numbering:
            section = section.replace(r'\subsubsection{', r'\subsubsection*{')
        else:
            section = section.replace(r'\subsubsection*{', r'\subsubsection{')

        if not self.allow_displayitem_tilde:
            section = section.replace(r'Table\textasciitilde', r'Table ')
            section = section.replace(r'Table \textasciitilde', r'Table ')
            section = section.replace(r'Figure\textasciitilde', r'Figure ')
            section = section.replace(r'Figure \textasciitilde', r'Figure ')

        section = evaluate_latex_num_command(section)[0]

        if self.replace_scientific_exponents:
            section = process_latex_text_and_math(
                section,
                process_text=partial(replace_scientific_exponent_with_latex, with_dollar_signs=True),
                process_math=partial(replace_scientific_exponent_with_latex, with_dollar_signs=False),
            )

        return section

    def get_document(self,
                     content: Optional[Union[str, Iterable[str], Dict[Optional[str], str]]] = None,
                     title: Optional[str] = None,
                     abstract: Optional[str] = None,
                     appendix: Optional[str] = None,
                     author: Optional[str] = None,
                     with_references: bool = False,
                     add_before_document: Optional[str] = None,
                     ) -> str:

        if isinstance(content, dict):
            if 'title' in content:
                title = content.pop('title')
            if 'abstract' in content:
                abstract = content.pop('abstract')

        # Build the document:
        s = ''
        s += r"\documentclass[{fontsize}pt]{{{kind}}}".format(kind=self.kind, fontsize=self.fontsize) + '\n'
        s += '\n'.join([r'\usepackage' + package for package in self.packages]) + '\n'

        s += '\\sectionfont{\\' + self.section_heading_fontsize + '}\n'
        s += '\\subsectionfont{\\' + self.subsection_heading_fontsize + '}\n'
        s += '\\subsubsectionfont{\\' + self.subsubsection_heading_fontsize + '}\n'

        s += '\n'.join(self.initiation_commands) + '\n'

        # Define title, author:
        if title is not None and not title.startswith(r'\title'):
            title = r'\title{' + title + '}'
        if title is not None:
            s += title + '\n'

        author = author or self.author
        if author is not None and not author.startswith(r'\author'):
            author = r'\author{' + author + '}'
        if author is not None:
            s += author + '\n'

        # Add before document:
        if add_before_document is not None:
            s += add_before_document + '\n'

        # Begin document:
        s += r'\begin{document}' + '\n'
        if title is not None:
            s += r'\maketitle' + '\n'

        # Abstract:
        if abstract is not None and not (abstract.startswith(r'\abstract') or abstract.startswith(r'\begin{abstract}')):
            abstract = r'\begin{abstract}' + abstract + r'\end{abstract}'
        if abstract is not None:
            s += abstract + '\n'

        # Content:
        if isinstance(content, str):
            all_sections = content
        elif isinstance(content, dict):
            all_sections = ''
            for section_name, section_content in content.items():
                if not section_content.startswith(r'\section') and not section_content.startswith(r'\subsection') and \
                        section_name is not None:
                    all_sections += r'\section{' + section_name + '}' + '\n'
                all_sections += section_content + '\n\n'
        elif isinstance(content, Iterable):
            all_sections = '\n\n'.join(content) + '\n'
        elif content is None:
            all_sections = ''
        else:
            raise TypeError(f'content must be str, Iterable[str] or Dict[str, str], not {type(content)}')

        s += self._style_section(all_sections)

        # References:
        if with_references:
            s += CITATION_TEMPLATE + '\n'

        # Appendix:
        if appendix is not None:
            s += START_APPENDIX + '\n'
            s += appendix + '\n\n'

        # End document:
        s += r'\end{document}' + '\n'

        return s

    def _get_is_pdflatex_installed(self) -> bool:
        global IS_PDFLATEX_INSTALLED
        if IS_PDFLATEX_INSTALLED is None:
            IS_PDFLATEX_INSTALLED = is_pdflatex_installed()
        return IS_PDFLATEX_INSTALLED

    def _get_missing_pdflatex_packages(self):
        global MISSING_PDFLATEX_PACKAGES
        if MISSING_PDFLATEX_PACKAGES is None:
            MISSING_PDFLATEX_PACKAGES = []
            for package_name in self.package_names:
                is_installed = is_pdflatex_package_installed(package_name)
                if is_installed is None:
                    raise MissingInstallationError(
                        package_name='pdflatex',
                        instructions='Failed checking pdflatex sub-packages. Please install pdflatex first.')
                status = 'Installed' if is_installed else 'Not installed'
                print(f'pdflatex package `{package_name}`: {status}')
                if not is_installed:
                    MISSING_PDFLATEX_PACKAGES.append(package_name)
        return MISSING_PDFLATEX_PACKAGES

    def raise_if_pdflatex_is_not_installed(self):
        """
        Check that pdflatex is installed.
        """
        if not self._get_is_pdflatex_installed():
            raise MissingInstallationError(package_name='pdflatex', instructions=PDFLATEX_INSTALLATION_INSTRUCTIONS)

    @property
    def package_names(self):
        return [package.split('{')[1].split('}')[0] for package in self.packages]

    def raise_if_packages_are_not_installed(self):
        """
        Check that the packages used in the latex document are installed.
        """
        missing_packages = self._get_missing_pdflatex_packages()
        if missing_packages:
            raise MissingInstallationError(package_name=f'pdflatex packages {missing_packages}',
                                           instructions='Please install the missing packages.')

    def compile_document(self,
                         content: Optional[Union[str, Iterable[str], Dict[Optional[str], str]]] = None,
                         title: Optional[str] = None,
                         abstract: Optional[str] = None,
                         appendix: Optional[str] = None,
                         author: Optional[str] = None,
                         references: Collection[Citation] = None,
                         format_cite: bool = True,
                         add_before_document: Optional[str] = None,
                         figures_folder: Optional[Path] = None,
                         file_stem: str = 'test',
                         output_directory: Optional[str] = None,
                         ) -> (str, str, Optional[float]):
        """
        Return the latex document as a string.

        If `file_stem` is given, save the document to a file and compile it to pdf.

        If `output_directory` is given, save the document to that directory.

        If `output_directory` is None:
            compile to pdf but do not save (checking for compilation errors)

        Returns:
            - The latex content.
            - The pdf compilation output.
            - The overflow width of the table, if any.

        `LatexCompilationError` is raised if there are errors.
        """

        self.raise_if_pdflatex_is_not_installed()
        self.raise_if_packages_are_not_installed()

        latex = self.get_document(content=content, title=title, abstract=abstract, appendix=appendix, author=author,
                                  with_references=bool(references), add_before_document=add_before_document)
        pdf_output, over_width_pts = save_latex_and_compile_to_pdf(latex, file_stem=file_stem,
                                                                   output_directory=output_directory,
                                                                   references=references,
                                                                   format_cite=format_cite,
                                                                   figures_folder=figures_folder)
        return latex, pdf_output, over_width_pts

    def compile_table(self, latex_table: str, file_stem: str = 'test', output_directory: Optional[str] = None) -> float:
        """
        Compile a latex table to pdf and return the width of the tabular part of the table,
        expressed as fraction of the page margin width.
        """

        lrbox_table = dedent_triple_quote_str(r"""
            % Define the save box within the document block
            \newsavebox{\mytablebox} % Create a box to store the table

            % Save only the tabular part of table in the \mytablebox without typesetting it:
            \begin{lrbox}{\mytablebox}
              <tabular>%
            \end{lrbox}

            % Typeset the entire table:
            <table>

            % Print the width of the tabular part of the table and the width of the page margin to the log file
            \typeout{Table width: \the\wd\mytablebox}
            \typeout{Page margin width: \the\textwidth}
            """).replace('<tabular>', get_tabular_block(latex_table)).replace('<table>', latex_table)

        _, pdf_output, over_width_pts = self.compile_document(
            content=lrbox_table, format_cite=False, file_stem=file_stem, output_directory=output_directory)

        table_width = re.findall(pattern=r'Table width: (\d+\.\d+)pt', string=pdf_output)[0]
        margin_width = re.findall(pattern=r'Page margin width: (\d+\.\d+)pt', string=pdf_output)[0]
        return float(table_width) / float(margin_width)



import importlib
import re
import subprocess

from data_to_paper.utils.subprocess_call import get_subprocess_kwargs
from data_to_paper.terminate.exceptions import MissingInstallationError
from data_to_paper.terminate.resource_checking import resource_checking
from data_to_paper.latex.clean_latex import process_latex_text_and_math
from data_to_paper.utils.file_utils import run_in_temp_directory
from data_to_paper.text.text_formatting import escape_html


@resource_checking("Checking Pandoc installation")
def check_pandoc_is_installed():
    # This will raise MissingInstallationError:
    raise_if_pandoc_is_not_installed()


def raise_if_pandoc_is_not_installed():
    try:
        subprocess.run(['pandoc', '--version'], **get_subprocess_kwargs(capture=False))
    except FileNotFoundError:
        raise MissingInstallationError(package_name="Pandoc", instructions="See: https://pandoc.org/installing.html")


def convert_latex_to_html(latex: str) -> str:
    """
    Convert LaTeX text to HTML using Pandoc through pypandoc.

    Parameters:
    - latex (str): A string containing LaTeX code.

    Returns:
    - str: The converted HTML text.
    """
    raise_if_pandoc_is_not_installed()

    is_title = re.search(pattern=r'\\title{(.+?)}', string=latex)

    tex_file = 'temp.tex'

    # Get the html template:
    if is_title:
        template_name = 'html_template_for_title_latex.html'
    else:
        template_name = 'html_template_for_titleless_latex.html'

    with importlib.resources.path('data_to_paper.latex.resources', 'adjust_section.lua') as lua_file, \
            importlib.resources.path('data_to_paper.latex.resources', template_name) as template_path:
        command = [
            'pandoc', '-s', tex_file,
            '--lua-filter', str(lua_file),
            '-t', 'html'
        ]

        command += ['--template', str(template_path)]

        if not is_title:
            command += ['--metadata', 'title=Titleless LaTeX Document']

        # To show citation commands (like '\\cite{ref1}'):
        command += ['--citeproc']

        try:
            with run_in_temp_directory():
                # process latex and escape special characters
                latex = process_latex_text_and_math(latex)
                # Write the LaTeX into a temporary file
                with open(tex_file, 'w', encoding='utf-8') as f:
                    f.write(latex)
                # Convert using Pandoc
                output = subprocess.run(command, universal_newlines=True, **get_subprocess_kwargs())
                return output.stdout
        except subprocess.CalledProcessError:
            # In case of an error, return the raw latex with proper escaping for HTML
            return escape_html(latex)



import functools
import os
import pickle
import time
from abc import ABC
from pathlib import Path
from typing import Union, Optional

from data_to_paper.env import CHOSEN_APP, DELAY_SERVER_CACHE_RETRIEVAL
from .json_dump import dump_to_json, load_from_json
from .serialize_exceptions import (
    serialize_exception,
    is_exception,
    de_serialize_exception,
)


class NoMoreResponsesToMockError(Exception):
    pass


def recursively_convert_lists_and_dicts_to_tuples(obj):
    if isinstance(obj, (list, tuple)):
        return tuple(
            recursively_convert_lists_and_dicts_to_tuples(item) for item in obj
        )
    if isinstance(obj, dict):
        return tuple(
            (
                recursively_convert_lists_and_dicts_to_tuples(key),
                recursively_convert_lists_and_dicts_to_tuples(value),
            )
            for key, value in obj.items()
        )
    return obj


def convert_args_kwargs_to_tuple(args, kwargs):
    return recursively_convert_lists_and_dicts_to_tuples((args, kwargs))


class ServerCaller(ABC):
    """
    A base class for calling a remote server, while allowing recording and replaying server responses.
    """

    name: str = None
    file_extension: str = None

    def __init__(self, fail_if_not_all_responses_used=True):

        self.old_records = self.empty_records
        self.new_records = self.empty_records
        self.args_kwargs_response_history = []  # for debugging
        self.is_playing_or_recording = False
        self.record_more_if_needed = False
        self.fail_if_not_all_responses_used = fail_if_not_all_responses_used
        self.should_save = False
        self.file_path = None

    @property
    def empty_records(self) -> Union[list, dict]:
        raise NotImplementedError()

    @property
    def all_records(self) -> Union[list, dict]:
        raise NotImplementedError()

    @classmethod
    def _get_server_response(cls, *args, **kwargs):
        """
        actual call to the server.
        this method should only return raw responses that can be serialized to json, without losing type information.
        """
        raise NotImplementedError()

    @staticmethod
    def _post_process_response(response, args, kwargs):
        """
        post process the response before transmitting.
        """
        return response

    def _save_records(self, records, filepath):
        """
        saves the records to a file.
        """
        raise NotImplementedError()

    def _load_records(self, filepath):
        """
        loads the records from a file.
        """
        raise NotImplementedError()

    def get_server_response(self, *args, **kwargs):
        """
        returns the response from the server after post-processing. allows recording and replaying.
        """
        response = self._get_raw_server_response(*args, **kwargs)
        if isinstance(response, Exception):
            raise response
        return self._post_process_response(response, args, kwargs)

    def _get_response_from_records(self, args, kwargs):
        """
        returns the response from the records, if exists.
        """
        raise NotImplementedError()

    def _add_response_to_new_records(self, args, kwargs, response):
        """
        adds a response to the new records.
        """
        raise NotImplementedError()

    def _get_raw_server_response(self, *args, **kwargs):
        """
        returns the raw response from the server, allows recording and replaying.
        """
        if not self.is_playing_or_recording:
            return self._get_server_response(*args, **kwargs)
        response = self._get_response_from_records(args, kwargs)
        if response is not None and CHOSEN_APP is not None:
            time.sleep(DELAY_SERVER_CACHE_RETRIEVAL.val)
        if response is None:
            if not self.record_more_if_needed:
                raise NoMoreResponsesToMockError()
            response = self._get_server_response(*args, **kwargs)
            self._add_response_to_new_records(args, kwargs, response)
            if self.should_save:
                self.save_records()
        self.args_kwargs_response_history.append(
            (args, kwargs, response)
        )  # for debugging and testing
        return response

    def __enter__(self):
        self.new_records = self.empty_records
        self.is_playing_or_recording = True
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.is_playing_or_recording = False
        if self.should_save:
            self.save_records()
        return False  # do not suppress exceptions

    def mock(
        self,
        old_records=None,
        record_more_if_needed=True,
        fail_if_not_all_responses_used=True,
        should_save=False,
        file_path=None,
    ):
        """
        Returns a context manager to mock the server responses (specified as old_records).
        """
        self.old_records = old_records or self.empty_records
        self.args_kwargs_response_history = []
        self.record_more_if_needed = record_more_if_needed
        self.fail_if_not_all_responses_used = fail_if_not_all_responses_used
        self.should_save = should_save
        self.file_path = file_path
        return self

    def save_records(self, file_path: Optional[str] = None):
        """
        Save the recorded responses to a file.
        """
        file_path = file_path or self.file_path
        # create the directory if not exist
        Path(os.path.dirname(file_path)).mkdir(parents=True, exist_ok=True)
        self._save_records(self.all_records, file_path)

    def mock_with_file(
        self,
        file_path,
        record_more_if_needed=True,
        fail_if_not_all_responses_used=True,
        should_save=True,
    ):
        """
        Returns a context-manager to mock the server responses from a specified file.
        """
        # load the old records from the file if exist
        if os.path.isfile(file_path):
            old_records = self._load_records(file_path)
        else:
            old_records = []

        return self.mock(
            old_records=old_records,
            record_more_if_needed=record_more_if_needed,
            fail_if_not_all_responses_used=fail_if_not_all_responses_used,
            should_save=should_save,
            file_path=file_path,
        )

    def record_or_replay(
        self,
        file_path: Union[str, Path] = None,
        should_mock: bool = True,
        fail_if_not_all_responses_used: bool = True,
    ):
        """
        Returns a decorator to call the decorated function while recording or replaying server responses.

        If the file exist, the responses will be replayed from the file.
        If the file does not exist, the responses will be recorded to the file.
        """

        def decorator(func):
            if not hasattr(func, "_module_file_path"):
                func._module_file_path = os.path.dirname(
                    os.path.abspath(func.__code__.co_filename)
                )

            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                nonlocal file_path
                # use the file path of the decorated function if not given
                file_path = file_path or os.path.join(
                    func._module_file_path,
                    "recorded_responses",
                    func.__name__ + self.file_extension,
                )

                # run the test with the previous responses and record new responses
                with self.mock_with_file(
                    file_path=file_path,
                    fail_if_not_all_responses_used=fail_if_not_all_responses_used,
                ):
                    func(*args, **kwargs)

            wrapper._module_file_path = func._module_file_path
            return wrapper if should_mock else func

        return decorator


class OrderedServerCaller(ServerCaller, ABC):
    """
    A base class for calling
    a remote server, while allowing recording and replaying server responses in the same order.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.index_in_old_records = 0

    def are_more_records_available(self):
        return self.index_in_old_records < len(self._get_old_records_as_list())

    def reset_index(self):
        self.index_in_old_records = 0

    def _get_old_records_as_list(self):
        """
        Return a single list of all the old records, as tuples of (key, value).
        """
        raise NotImplementedError()

    def _get_response_from_a_record(self, record, args, kwargs):
        return record

    def _get_response_from_records(self, args, kwargs):
        if self.are_more_records_available():
            record = self._get_old_records_as_list()[self.index_in_old_records]
            response = self._get_response_from_a_record(record, args, kwargs)
            self.index_in_old_records += 1
            return response
        return None

    @staticmethod
    def _serialize_record(record):
        if isinstance(record, Exception):
            return serialize_exception(record)
        return record

    @staticmethod
    def _deserialize_record(serialized_record):
        if is_exception(serialized_record):
            return de_serialize_exception(serialized_record)
        return serialized_record

    def __exit__(self, exc_type, exc_val, exc_tb):
        results = super().__exit__(exc_type, exc_val, exc_tb)
        if self.fail_if_not_all_responses_used and self.are_more_records_available():
            raise AssertionError(
                f"Not all responses were used ({self.__class__.__name__})."
            )
        return results

    def __enter__(self):
        self.reset_index()
        return super().__enter__()


class ListServerCaller(OrderedServerCaller):
    """
    A base class for calling a remote server, while allowing recording and replaying server responses.
    Records are saved as a sequence of responses and can be replayed in the same order (regardless of the arguments).
    """

    @property
    def empty_records(self) -> list:
        return []

    @property
    def all_records(self):
        return self.old_records + self.new_records

    def _get_old_records_as_list(self):
        return self.old_records

    def _add_response_to_new_records(self, args, kwargs, response):
        self.new_records.append(response)

    def _save_records(self, records, filepath):
        dump_to_json([self._serialize_record(record) for record in records], filepath)

    def _load_records(self, filepath):
        return [
            self._deserialize_record(serialized_record)
            for serialized_record in load_from_json(filepath)
        ]


class OrderedKeyToListServerCaller(OrderedServerCaller):
    """
    A class for calling a remote server, while allowing recording and replaying server responses.
    Records are saved as dictionary (key order preserving) of responses with ordered lists as values.
    """

    @property
    def empty_records(self) -> dict:
        return {}

    @property
    def all_records(self):
        records = self.old_records.copy()
        for key, values in self.new_records.items():
            records[key] = records.get(key, []) + values
        return records

    def _get_old_records_as_list(self):
        """
        Return a single list of all the old records, as tuples of (key, value).
        """
        return [
            (key, value) for key, values in self.old_records.items() for value in values
        ]

    def _get_response_from_a_record(self, record, args, kwargs):
        # record is (key, value)
        key = self._generate_key(args, kwargs)
        if not key == record[0]:
            raise ValueError(f"Key mismatch: {key} != {record[0]}")
        return record[1]

    def _add_response_to_new_records(self, args, kwargs, response):
        key = self._generate_key(args, kwargs)
        if key not in self.new_records:
            self.new_records[key] = []
        self.new_records[key].append(response)

    def _generate_key(self, args, kwargs):
        return convert_args_kwargs_to_tuple(args, kwargs)

    def _save_records(self, records, filepath):
        serialized_records = {
            key: [self._serialize_record(record) for record in value]
            for key, value in records.items()
        }
        dump_to_json(serialized_records, filepath)

    def _load_records(self, filepath):
        serialized_records = load_from_json(filepath)
        return {
            key: [self._deserialize_record(record) for record in value]
            for key, value in serialized_records.items()
        }

    def mock(self, old_records=None, *args, **kwargs):
        """
        Returns a context manager to mock the server responses (specified as old_records).
        """
        result = super().mock(old_records, *args, **kwargs)
        self.old_records = (
            old_records
            if isinstance(old_records, dict)
            else {"GENERAL": old_records} if (old_records) else self.empty_records
        )
        return result


class ParameterizedQueryServerCaller(ServerCaller, ABC):
    """
    A base class for calling a remote server, while allowing recording and replaying server responses.
    Records are saved as a dictionary of responses and can be replayed by the arguments and keyword arguments.
    """

    @property
    def empty_records(self) -> dict:
        return {}

    @property
    def all_records(self):
        return self.old_records | self.new_records

    def _get_response_from_records(self, args, kwargs):
        tuple_args_and_kwargs = convert_args_kwargs_to_tuple(args, kwargs)
        return self.all_records.get(tuple_args_and_kwargs, None)

    def _add_response_to_new_records(self, args, kwargs, response):
        tuple_args_and_kwargs = convert_args_kwargs_to_tuple(args, kwargs)
        self.new_records[tuple_args_and_kwargs] = response

    def _save_records(self, records, filepath):
        with open(filepath, "wb") as file:
            pickle.dump(records, file)

    def _load_records(self, filepath):
        with open(filepath, "rb") as filepath:
            return pickle.load(filepath)



import os
import re

from pathlib import Path
from ansi2html import Ansi2HTMLConverter


def convert_ansi_to_html(ansi_text):
    conv = Ansi2HTMLConverter()
    html_text = conv.convert(ansi_text, full=True)
    return html_text


def filter_text(text):
    lines = text.split('\n')
    lines_to_filter_startswith = [
        ' [31mERROR: None embedding attr.',
        ' [31mCreateConversation(',
    ]
    # Filter out lines based on the startswith conditions
    filtered_lines = [
        line for line in lines
        if not any(line.startswith(prefix) for prefix in lines_to_filter_startswith)
    ]
    # Remove everything from "This is BibTeX," to the end of the document
    for i, line in enumerate(filtered_lines):
        if line.startswith("This is BibTeX,"):
            filtered_lines = filtered_lines[:i]
            break
    # Join the lines into a single string
    filtered_text = '\n'.join(filtered_lines)
    # Replace three or more consecutive newline characters with two newline characters
    filtered_text = re.sub(r'\n{3,}', '\n', filtered_text)
    return filtered_text


def convert_console_log_to_html(console_filepath: Path):
    """
    Convert the console log to a html file.
    """
    # check if file exists and is not empty
    if not os.path.isfile(console_filepath) or not os.path.getsize(console_filepath) > 0:
        raise FileNotFoundError(f'File {console_filepath} does not exist or is empty')
    with open(console_filepath, 'r', encoding='utf-8') as f:
        text_as_string = f.read()
        filtered_text = filter_text(text_as_string)
        text_as_string_html = convert_ansi_to_html(filtered_text)
        html_file = console_filepath.parent / (console_filepath.stem + '.html')
        with open(html_file, 'w', encoding='utf-8') as new_f:
            new_f.write(text_as_string_html)
    return html_file



import json


def dump_to_json(serialized_data, filename):
    with open(filename, 'w') as file:
        json.dump(serialized_data, file, indent=4, sort_keys=False)


def load_from_json(filename):
    with open(filename, 'r') as file:
        return json.load(file)



from dataclasses import dataclass
from typing import Tuple, Collection, Optional

from data_to_paper.base_steps import BaseLiteratureSearchReviewGPT
from data_to_paper.research_types.hypothesis_testing.cast import ScientificAgent
from data_to_paper.research_types.hypothesis_testing.scientific_products import ScientificProducts


@dataclass
class GoalLiteratureSearchReviewGPT(BaseLiteratureSearchReviewGPT):
    products: ScientificProducts = None
    requested_keys: Collection[str] = ('dataset', 'questions', )
    background_product_fields: Tuple[str, ...] = ('data_file_descriptions_no_headers', 'research_goal')
    conversation_name: str = 'Literature Search for Goal'
    is_new_conversation: bool = None
    goal_noun: str = 'literature search queries'
    goal_verb: str = 'write'
    assistant_agent: ScientificAgent = ScientificAgent.Performer
    user_agent: ScientificAgent = ScientificAgent.CitationExpert


@dataclass
class WritingLiteratureSearchReviewGPT(GoalLiteratureSearchReviewGPT):
    requested_keys: Collection[str] = ('background', 'dataset', 'methods', 'results')
    background_product_fields: Tuple[str, ...] = ('data_file_descriptions_no_headers', 'research_goal',
                                                  'hypothesis_testing_plan', 'title_and_abstract')
    conversation_name: str = 'Literature Search for Writing'

    def get_abstract(self) -> Optional[str]:
        return self.products.get_abstract()

    def get_title(self) -> Optional[str]:
        return self.products.get_title()



from __future__ import annotations
import re
from dataclasses import dataclass
from enum import Enum
from typing import Tuple, List, Optional

TARGET = r'\hypertarget'
LINK = r'\hyperlink'


def get_numeric_value_pattern(must_follow: Optional[str] = None, allow_commas: bool = True) -> str:
    """
    Get a pattern for a numeric value that must follow a sequence of characters.
    """
    # if must_follow is None:
    #     must_follow = ''
    # if allow_commas:
    #     return fr'(?<={must_follow})(?:[-+]?\d+(?:,\d{{3}})*(?:\.\d+)?(?:e[-+]?\d+)?|\d{{1,3}}(?:,\d{{3}})+)(?!\d)'
    # return fr'(?<={must_follow})[+-]?\d+(?:\.\d+)?(?:[eE][+-]?\d+)?'

    if must_follow is None:
        must_follow = ''
    else:
        must_follow = fr'(?<={must_follow})'
    if allow_commas:
        pattern = r'(?:[-+]?\d+(?:,\d{3})*(?:\.\d+)?(?:e[-+]?\d+)?|\d{1,3}(?:,\d{{3}})+)(?!\d)'
    else:
        pattern = r'[-+]?\d+(?:\.\d+)?(?:[eE][+-]?\d+)?'
    return must_follow + pattern


numeric_values_in_products_pattern = get_numeric_value_pattern(must_follow=r'[$,{<=\s\n\(\[]', allow_commas=True)
numeric_values_in_response_pattern = get_numeric_value_pattern(must_follow=r'[$,{<=\s\n\(\[]', allow_commas=True)


def get_hyperlink_pattern(is_target: bool = False) -> str:
    """
    Get a pattern for a hyperlink.
    """
    command = TARGET if is_target else LINK
    return fr'\{command}{{(?P<reference>[^}}]*)}}{{(?P<value>[^}}]*)}}'


@dataclass
class ReferencedValue:
    """
    A numeric value with a reference to it.
    """
    value: str
    label: Optional[str] = None  # if None for un-referenced numeric values
    is_target: bool = True

    @property
    def command(self) -> str:
        return TARGET if self.is_target else LINK

    def to_str(self, hypertarget_format: Optional[HypertargetFormat] = None) -> str:
        if hypertarget_format is None:
            hypertarget_format = HypertargetFormat(position=HypertargetPosition.WRAP)
        value = self.value
        if hypertarget_format.position == HypertargetPosition.NONE or self.label is None:
            return value
        target = fr'{self.command}{{{self.label}}}'
        if hypertarget_format.position == HypertargetPosition.WRAP:
            target += fr'{{{value}}}'
            value = ''
        elif hypertarget_format.position == HypertargetPosition.ADJACENT:
            target += fr'{{}}'
        elif hypertarget_format.position == HypertargetPosition.HEADER:
            target += fr'{{}}'
            value = ''
        else:
            raise ValueError(f'Invalid hypertarget position: {hypertarget_format.position}')
        if hypertarget_format.raised:
            target = fr'\raisebox{{2ex}}{{{target}}}'
        if hypertarget_format.escaped:
            target = fr'(*@{target}@*)'
        return target + value

    def __str__(self) -> str:
        return self.to_str()

    def get_numeric_value_and_is_percent(self) -> Tuple[Optional[str], bool]:
        """
        Get the numeric value from a string.
        """
        value = self.value.strip()
        pattern = get_numeric_value_pattern()
        if re.fullmatch(pattern, value.strip()):
            return value, False
        for add in [r'\%', '%', r' \%', ' %']:
            if bool(re.fullmatch(pattern + re.escape(add), value)):
                return value[:-(len(add))], True
        return None, False

    def to_float(self) -> Optional[float]:
        value, is_percent = self.get_numeric_value_and_is_percent()
        if value is None:
            return None
        value = float(value.strip().replace(',', ''))
        return value / 100 if is_percent else value


class HypertargetPosition(Enum):
    NONE = 0  # do not create hypertargets
    WRAP = 1  # \hypertarget{target}{value}
    ADJACENT = 2  # \hypertarget{target}{}value
    HEADER = 3  # \hypertarget{target}{} ... value

    def __bool__(self):
        return self != HypertargetPosition.NONE


@dataclass(frozen=True)
class HypertargetFormat:
    position: HypertargetPosition = HypertargetPosition.NONE
    raised: bool = False  # \raisebox{2ex}{...}
    escaped: bool = False  # (*@ ... @*)

    def __bool__(self):
        return bool(self.position)

    def is_hypertarget_position_header(self) -> bool:
        return self.position == HypertargetPosition.HEADER


def find_hyperlinks(text: str, is_targets: bool = False) -> List[ReferencedValue]:
    """
    Find all references in the text.
    """
    pattern = get_hyperlink_pattern(is_targets)
    references = []
    for match in re.finditer(pattern, text):
        references.append(ReferencedValue(value=match.group('value'), label=match.group('reference'),
                                          is_target=False))
    return references


def replace_hyperlinks_with_values(text: str, is_targets: bool = False) -> str:
    """
    Replace all references in the text with the referenced values.
    """
    def replace_hyperlink_with_value(match):
        return match.group('value')

    pattern = get_hyperlink_pattern(is_targets)
    return re.sub(pattern, replace_hyperlink_with_value, text)


def find_numeric_values(text: str, remove_hyperlinks: bool = True,
                        pattern: str = numeric_values_in_response_pattern,
                        ) -> List[str]:
    """
    Find all numeric values in the text that are not referenced.
    """
    text = ' ' + text + ' '
    # remove all references from the text
    if remove_hyperlinks:
        text = re.sub(get_hyperlink_pattern(is_target=False), '', text)
    # find all numeric values
    return re.findall(pattern, text)


def find_matching_reference(query: ReferencedValue,
                            references: List[ReferencedValue]) -> Optional[ReferencedValue]:
    """
    Find the reference in references that matches the query.
    We assume that there is a unique match.
    Return None if no match was found.
    """
    for reference in references:
        if query.label == reference.label:
            return reference
    return None



# pragma: no cover
"""
This file is obsolete. It was used to create a custom p-value dtype for statsmodels to be used with
PValueFloat which inherits from float.
We now use a PValue object class instead of the PValueFloat(float) class.
"""


from __future__ import annotations

import operator
import re
from typing import Any, Sequence

import numpy as np
import pandas as pd
from pandas.core.dtypes.dtypes import PandasExtensionDtype
from pandas.core.indexers import check_array_indexer


class PValueFloat(float):
    pass


# Based on:
# https://stackoverflow.com/questions/68893521/simple-example-of-pandas-extensionarray:

@pd.api.extensions.register_extension_dtype
class PValueDtype(PandasExtensionDtype):
    """
    An ExtensionDtype for func-aware angular data.
    """
    # Required for all parameterized dtypes
    _metadata = ('func',)
    _match = re.compile(pattern=r'(P|p)value\[(?P<func>.+)\]')
    name = 'pvalue'

    def __init__(self, func=None):
        self._func = func

    def __str__(self) -> str:
        return f'pvalue[{self.func}]'

    # TestDtypeTests
    def __hash__(self) -> int:
        return hash(str(self))

    # TestDtypeTests
    def __eq__(self, other: Any) -> bool:
        if isinstance(other, str):
            return self.name == other
        else:
            return isinstance(other, type(self)) and self.func == other.func

    # Required for pickle compat (see GH26067)
    def __setstate__(self, state) -> None:
        self._func = state['func']

    # Required for all ExtensionDtype subclasses
    @classmethod
    def construct_array_type(cls):
        """
        Return the array type associated with this dtype.
        """
        return PValueArray

    # Required for all ExtensionDtype subclasses
    @property
    def type(self):
        """
        The scalar type for the array (e.g., int).
        """
        return np.float64

    @property
    def func(self) -> str:
        """
        The function used to compute the p-value.
        """
        return self._func


class PValueArray(pd.api.extensions.ExtensionArray):
    """
    An ExtensionArray for func-aware p-values.
    """

    # Include `copy` param for TestInterfaceTests
    def __init__(self, data, func=None, copy: bool = False):
        self._data = np.array(data, copy=copy)
        self._func = func

    # Required for all ExtensionArray subclasses
    def __getitem__(self, index: int) -> PValueArray | Any:
        """
        Select a subset of self.
        """
        if isinstance(index, int):
            return PValueFloat(self._data[index])
        else:
            # Check index for TestGetitemTests
            index = check_array_indexer(self, index)
            return type(self)(self._data[index])

    # TestSetitemTests
    def __setitem__(self, index: int, value: np.generic) -> None:
        """
        Set one or more values in-place.
        """
        # Check index for TestSetitemTests
        index = check_array_indexer(self, index)

        # Upcast to value's type (if needed) for TestMethodsTests
        if self._data.dtype < type(value):
            self._data = self._data.astype(type(value))

        # TODO: Validate value for TestSetitemTests
        # value = self._validate_setitem_value(value)

        self._data[index] = value

    # Required for all ExtensionArray subclasses
    def __len__(self) -> int:
        """
        Length of this array.
        """
        return len(self._data)

    # TestUnaryOpsTests
    def __invert__(self) -> PValueArray:
        """
        Element-wise inverse of this array.
        """
        data = ~self._data
        return type(self)(data, func=self.dtype.func)

    def _apply_operator(self, op, other, recast=False) -> np.ndarray | PValueArray:
        """
        Helper method to apply an operator `op` between `self` and `other`.

        Some ops require the result to be recast into PValueArray:
        * Comparison ops: recast=False
        * Arithmetic ops: recast=True
        """
        f = operator.attrgetter(op)
        data, other = np.array(self), np.array(other)
        result = f(data)(other)
        return result if not recast else type(self)(result, func=self.dtype.func)

    def _apply_operator_if_not_series(self, op, other, recast=False) -> np.ndarray | PValueArray:
        """
        Wraps _apply_operator only if `other` is not Series/DataFrame.

        Some ops should return NotImplemented if `other` is a Series/DataFrame:
        https://github.com/pandas-dev/pandas/blob/e7e7b40722e421ef7e519c645d851452c70a7b7c/pandas/tests/extension/base/ops.py#L115
        """
        if isinstance(other, (pd.Series, pd.DataFrame)):
            return NotImplemented
        else:
            return self._apply_operator(op, other, recast=recast)

    # Required for all ExtensionArray subclasses
    @pd.core.ops.unpack_zerodim_and_defer('__eq__')
    def __eq__(self, other):
        return self._apply_operator('__eq__', other, recast=False)

    # TestComparisonOpsTests
    @pd.core.ops.unpack_zerodim_and_defer('__ne__')
    def __ne__(self, other):
        return self._apply_operator('__ne__', other, recast=False)

    # TestComparisonOpsTests
    @pd.core.ops.unpack_zerodim_and_defer('__lt__')
    def __lt__(self, other):
        return self._apply_operator('__lt__', other, recast=False)

    # TestComparisonOpsTests
    @pd.core.ops.unpack_zerodim_and_defer('__gt__')
    def __gt__(self, other):
        return self._apply_operator('__gt__', other, recast=False)

    # TestComparisonOpsTests
    @pd.core.ops.unpack_zerodim_and_defer('__le__')
    def __le__(self, other):
        return self._apply_operator('__le__', other, recast=False)

    # TestComparisonOpsTests
    @pd.core.ops.unpack_zerodim_and_defer('__ge__')
    def __ge__(self, other):
        return self._apply_operator('__ge__', other, recast=False)

    # TestArithmeticOpsTests
    @pd.core.ops.unpack_zerodim_and_defer('__add__')
    def __add__(self, other) -> PValueArray:
        return self._apply_operator_if_not_series('__add__', other, recast=True)

    # TestArithmeticOpsTests
    @pd.core.ops.unpack_zerodim_and_defer('__sub__')
    def __sub__(self, other) -> PValueArray:
        return self._apply_operator_if_not_series('__sub__', other, recast=True)

    # TestArithmeticOpsTests
    @pd.core.ops.unpack_zerodim_and_defer('__mul__')
    def __mul__(self, other) -> PValueArray:
        return self._apply_operator_if_not_series('__mul__', other, recast=True)

    # TestArithmeticOpsTests
    @pd.core.ops.unpack_zerodim_and_defer('__truediv__')
    def __truediv__(self, other) -> PValueArray:
        return self._apply_operator_if_not_series('__truediv__', other, recast=True)

    # Required for all ExtensionArray subclasses
    @classmethod
    def _from_sequence(cls, data, dtype=None, copy: bool = False):
        """
        Construct a new PValueArray from a sequence of scalars.
        """
        if dtype is None:
            dtype = PValueDtype()

        if not isinstance(dtype, PValueDtype):
            msg = f"'{cls.__name__}' only supports 'PValueDtype' dtype"
            raise ValueError(msg)
        else:
            return cls(data, func=dtype.func, copy=copy)

    # TestParsingTests
    @classmethod
    def _from_sequence_of_strings(cls, strings, *, dtype=None, copy: bool = False) -> PValueArray:
        """
        Construct a new PValueArray from a sequence of strings.
        """
        scalars = pd.to_numeric(strings, errors='raise')
        return cls._from_sequence(scalars, dtype=dtype, copy=copy)

    # Required for all ExtensionArray subclasses
    @classmethod
    def _from_factorized(cls, uniques: np.ndarray, original: PValueArray):
        """
        Reconstruct an PValueArray after factorization.
        """
        return cls(uniques, func=original.dtype.func)

    # Required for all ExtensionArray subclasses
    @classmethod
    def _concat_same_type(cls, to_concat: Sequence[PValueArray]) -> PValueArray:
        """
        Concatenate multiple AngleArrays.
        """
        return cls(np.concatenate(to_concat), func=None)

    # Required for all ExtensionArray subclasses
    @property
    def dtype(self):
        """
        An instance of PValueDtype.
        """
        return PValueDtype(self._func)

    # Required for all ExtensionArray subclasses
    @property
    def nbytes(self) -> int:
        """
        The number of bytes needed to store this object in memory.
        """
        return self._data.nbytes

    @property
    def func(self):
        return self.dtype.func

    # Test*ReduceTests
    def all(self) -> bool:
        return all(self)

    def any(self) -> bool:  # Test*ReduceTests
        return any(self)

    def sum(self) -> np.generic:  # Test*ReduceTests
        return self._data.sum()

    def mean(self) -> np.generic:  # Test*ReduceTests
        return self._data.mean()

    def max(self) -> np.generic:  # Test*ReduceTests
        return self._data.max()

    def min(self) -> np.generic:  # Test*ReduceTests
        return self._data.min()

    def prod(self) -> np.generic:  # Test*ReduceTests
        return self._data.prod()

    def std(self) -> np.generic:  # Test*ReduceTests
        return pd.Series(self._data).std()

    def var(self) -> np.generic:  # Test*ReduceTests
        return pd.Series(self._data).var()

    def median(self) -> np.generic:  # Test*ReduceTests
        return np.median(self._data)

    def skew(self) -> np.generic:  # Test*ReduceTests
        return pd.Series(self._data).skew()

    def kurt(self) -> np.generic:  # Test*ReduceTests
        return pd.Series(self._data).kurt()

    # Test*ReduceTests
    def _reduce(self, name: str, *, skipna: bool = True, **kwargs):
        """
        Return a scalar result of performing the reduction operation.
        """
        f = operator.attrgetter(name)
        return f(self)()

    # Required for all ExtensionArray subclasses
    def isna(self):
        """
        A 1-D array indicating if each value is missing.
        """
        return pd.isnull(self._data)

    # Required for all ExtensionArray subclasses
    def copy(self):
        """
        Return a copy of the array.
        """
        copied = self._data.copy()
        return type(self)(copied, func=self.func)

    # Required for all ExtensionArray subclasses
    def take(self, indices, allow_fill=False, fill_value=None):
        """
        Take elements from an array.
        """
        if allow_fill and fill_value is None:
            fill_value = self.dtype.na_value

        result = pd.core.algorithms.take(self._data, indices, allow_fill=allow_fill,
                                         fill_value=fill_value)
        return self._from_sequence(result)

    # TestMethodsTests
    def value_counts(self, dropna: bool = True):
        """
        Return a Series containing descending counts of unique values (excludes NA values by default).
        """
        return pd.core.algorithms.value_counts(self._data, dropna=dropna)



from typing import List, Dict

from data_to_paper.utils.tag_pairs import TagPairs

SECTIONS_OR_FRAGMENTS_TO_TAG_PAIR_OPTIONS: Dict[str, List[TagPairs]] = {
    'title': [TagPairs('\\title{', '}')],
    'abstract': [TagPairs('\\begin{abstract}', '\\end{abstract}')],
    'table': [TagPairs('\\begin{table}', '\\end{table}')]
}


def get_list_of_tag_pairs_for_section_or_fragment(section_or_fragment: str) -> List[TagPairs]:
    """
    Get a list with different options for tag pairs enclosing the specified section.
    """
    if section_or_fragment in SECTIONS_OR_FRAGMENTS_TO_TAG_PAIR_OPTIONS:
        return SECTIONS_OR_FRAGMENTS_TO_TAG_PAIR_OPTIONS[section_or_fragment]
    else:
        return [TagPairs(f'\\section{{{section_or_fragment.capitalize()}}}', None),
                TagPairs(f'\\section*{{{section_or_fragment.capitalize()}}}', None)]



from dataclasses import dataclass, field
from typing import Tuple, Union, Any, Optional

from data_to_paper.text.text_extractors import extract_all_external_brackets
from data_to_paper.text.text_formatting import forgiving_format
from data_to_paper.utils.types import ListBasedSet


@dataclass
class Replacer:
    """
    A class to replace placeholders in a string with the attributes of a given object (or of multiple objects).
    Also allows adding additional args and kwargs for the string formatting.
    Formatting is done by calling forgiving_format(), so matching placeholders are replaced and non-matching
    are left as is.
    """
    objs: Optional[Union[Any, list]] = None
    text: str = ''
    args: Tuple = field(default_factory=tuple)
    kwargs: dict = field(default_factory=dict)

    def __str__(self):
        return self.format_text()

    def __bool__(self):
        return bool(self.text)

    def add_obj(self, obj):
        if self.objs is None:
            self.objs = obj
        elif not isinstance(self.objs, list):
            self.objs = [self.objs] + [obj]
        else:
            self.objs.append(obj)

    def get_objs(self):
        if self.objs is None:
            return []
        elif isinstance(self.objs, list):
            return self.objs
        else:
            return [self.objs]

    def format_text(self) -> str:
        text = self.text
        brackets = ListBasedSet(extract_all_external_brackets(self.text, '{'))
        additional_kwargs = {}
        objs = self.get_objs()
        for bracket in brackets:
            bracketed_text = bracket[1:-1]
            for obj in objs:
                if hasattr(obj, bracketed_text):
                    attr = getattr(obj, bracketed_text)
                    if not isinstance(attr, Replacer):
                        attr = Replacer(obj, str(attr))
                    attr = attr.format_text()
                    additional_kwargs[bracketed_text] = attr
                    break
            else:
                pass  # we don't have the attribute in any of the objects, so we don't do anything
        # add object kwargs:
        for obj in objs:
            if hasattr(obj, 'replacer_kwargs'):
                additional_kwargs.update(obj.replacer_kwargs)

        return forgiving_format(text, *self.args, **self.kwargs, **additional_kwargs)


def format_value(obj: object, value: Any, should_format: bool = True) -> Union[str, Any]:
    if not should_format:
        return value
    if isinstance(value, Replacer):
        value.add_obj(obj)
        return value.format_text()
    elif isinstance(value, str):
        return Replacer(obj, value).format_text()
    else:
        return value


StrOrReplacer = Union[str, Replacer]



import re
from dataclasses import dataclass, field
from typing import Dict, Tuple, List

from data_to_paper.text import line_count, dedent_triple_quote_str
from data_to_paper.run_gpt_code.code_utils import extract_code_from_text
from data_to_paper.run_gpt_code.run_issues import RunIssue, CodeProblem


def get_issue_for_use_of_a_forbidden_function(func: str, suggest_print_to_output: bool) -> RunIssue:
    category = 'Use of un-allowed functions'

    instructions = f"Do not use the function `{func}` in your code."
    if func == 'print':
        instructions += \
            "\nIf you print conditional warning messages, please use `assert` or `raise` instead."
        if suggest_print_to_output:
            instructions += \
                "\nOtherwise, outputs should only be written to the above described output file(s)."
    return RunIssue(
        category=category,
        issue=f"Your code uses the function `{func}`, which is not allowed.",
        instructions=instructions,
        code_problem=CodeProblem.RuntimeError,
        comment=f'Code uses forbidden function {func}',
    )


@dataclass
class CodeExtractor:
    """
    Extract code from a response.
    Keep track of line numbers added in front of the code.
    """

    def _get_raw_code(self, response) -> str:
        """
        Extract the raw code from the response.
        """
        return extract_code_from_text(response)

    def get_modified_code_and_num_added_lines(self, response) -> Tuple[str, int]:
        return self._get_raw_code(response), 0

    def get_issues_for_static_code_check(self, code: str) -> List[RunIssue]:
        return []


@dataclass
class ModifyAndCheckCodeExtractor(CodeExtractor):
    """
    Add additional code in front of the extracted code.
    Make replacements in the extracted code.
    """
    add_in_front_of_code: str = ''
    code_replacements: Dict[str, str] = field(default_factory=dict)

    headers_required_in_code: Tuple[str, ...] = ()
    phrases_required_in_code: Tuple[str, ...] = ()
    un_allowed_phrases: Tuple[str, ...] = ('__name__',)
    suggest_print_to_output: bool = False

    def get_modified_code_and_num_added_lines(self, response) -> Tuple[str, int]:
        """
        Modify the extracted code before running it.
        """
        code, num_added_lines = super().get_modified_code_and_num_added_lines(response)
        code = self.add_in_front_of_code + code
        num_added_lines += line_count(self.add_in_front_of_code)
        for old, new in self.code_replacements.items():
            code = code.replace(old, new)
        return code, num_added_lines

    def get_issues_for_static_code_check(self, code: str) -> List[RunIssue]:
        issues = []
        required_strings_not_found = [s for s in self.headers_required_in_code if s.lower() not in code.lower()]
        if len(required_strings_not_found) > 0:
            issues.append(RunIssue(
                category='Code structure',
                issue=dedent_triple_quote_str("""
                Your code must contain the following sections: 
                {headers_required_in_code}.
                But I could not find these headers:
                {required_strings_not_found}.
                """).format(
                    headers_required_in_code=self.headers_required_in_code,
                    required_strings_not_found=required_strings_not_found,
                ),
                instructions='Please rewrite the complete code again with all the required sections.',
                comment='Required sections not found',
                code_problem=CodeProblem.StaticCheck,
            ))

        # check if code uses `print`:
        if re.search(pattern=r'\bprint\s*\(', string=code):
            issues.append(get_issue_for_use_of_a_forbidden_function('print', self.suggest_print_to_output))

        # check if code has un-allowed keywords:
        for phrase in self.un_allowed_phrases:
            if phrase in code:
                issues.append(RunIssue(
                    category='Un-allowed phrases in code',
                    issue=f"Your code uses `{phrase}`, which is not allowed.",
                    instructions=f"Please rewrite the complete code again without using `{phrase}`.",
                    comment=f'Code uses forbidden phrase.',
                    code_problem=CodeProblem.StaticCheck,
                ))

        # check if code has all required keywords:
        for phrase in self.phrases_required_in_code:
            if phrase not in code:
                issues.append(RunIssue(
                    category='Missing/imperfect commands',
                    issue=f"Your code must explicitly use:\n`{phrase.strip()}`.",
                    comment=f'Code does not use required phrase.',
                    code_problem=CodeProblem.StaticCheck,
                ))

        return issues



import sys
from typing import Type, Optional, Union

from pathlib import Path

from data_to_paper.base_steps import BaseStepsRunner
from data_to_paper.env import CHOSEN_APP
from data_to_paper.interactive.base_app_startup import interactively_create_project_folder, BASE_PROJECT_DIRECTORY
from data_to_paper.interactive.get_app import create_app


def run_all_steps(step_runner: BaseStepsRunner):
    app = create_app(step_runner=step_runner)
    if app is None:
        step_runner.run_all_steps()
    else:
        step_runner.app = app
        exit_code = app.initialize()
        sys.exit(exit_code)


def set_project_and_run(steps_runner_cls: Type[BaseStepsRunner],
                        project_directory: Optional[Union[Path, str]] = None,
                        run_name: str = 'run_001'):
    if isinstance(project_directory, str):
        project_directory = Path(project_directory)
    if project_directory is not None and not project_directory.is_absolute():
        project_directory = BASE_PROJECT_DIRECTORY / project_directory
    if CHOSEN_APP == 'pyside':
        project_directory, config = interactively_create_project_folder(steps_runner_cls, project_directory)
        if project_directory is None:
            return
    else:
        if project_directory is None:
            raise ValueError("You must provide a project directory when not using the interactive app")
    if isinstance(run_name, str):
        output_directory = project_directory / 'runs' / run_name
    else:
        output_directory = run_name
    step_runner = steps_runner_cls(
        project_directory=project_directory,
        output_directory=output_directory,
    )
    run_all_steps(step_runner=step_runner)



from typing import Dict, Any, Union, List

from data_to_paper.utils.types import IndexOrderedEnum


class Stage(IndexOrderedEnum):
    """
    Define the sequence of stages
    Each stage is defined as (str, bool),
    indicating the name of the stage and whether the user
    can reset to this stage.
    """

    def __new__(cls, value, resettable):
        obj = object.__new__(cls)
        obj._value_ = value
        obj.resettable = resettable
        return obj


def get_all_keys_following_stage(data: Dict[Union[Stage, str], Any], stage: Stage,
                                 include_stage: bool = True
                                 ) -> List[Union[Stage, str]]:
    """
    Get all the keys of data that are following the given stage
    keys can be either Stage or str
    """
    stage_cls = stage.__class__
    keys = []
    for key in list(data.keys()):
        if isinstance(key, str):
            for stage_name in stage_cls:
                if stage_name.value == key:
                    current_stage = stage_name
                    break
            else:
                raise ValueError(f"Stage {key} not found in {stage_cls}")
        else:
            current_stage = key
        if (isinstance(current_stage, stage_cls) and
                (current_stage > stage or (include_stage and current_stage == stage))):
            keys.append(key)
    return keys


def delete_all_stages_following_stage(data: Dict[Union[Stage, str], Any], delete_from_stage: Stage,
                                      include_stage: bool = True):
    """
    Delete all stages following the given stage
    """
    keys = get_all_keys_following_stage(data, delete_from_stage, include_stage)
    for key in keys:
        del data[key]



# DO NOT DELETE PACKAGE
#
# This is where CodeRunner dynamically creates and run modules from LLM response.
# Any files here can be deleted, except for the folder and the __init__ file which must be kept.



from __future__ import annotations

from dataclasses import dataclass, field

from typing import Optional, Any, Union

from data_to_paper import Message
from data_to_paper.conversation.actions_and_conversations import ActionsAndConversations
from data_to_paper.env import TEXT_WIDTH
from data_to_paper.conversation import ConversationManager, GeneralMessageDesignation
from data_to_paper.interactive import PanelNames
from data_to_paper.interactive.app_interactor import AppInteractor, _raise_if_reset
from data_to_paper.servers.model_engine import ModelEngine
from data_to_paper.utils.replacer import StrOrReplacer, format_value
from data_to_paper.utils.print_to_file import print_and_log_red, print_and_log_magenta
from data_to_paper.base_cast import Agent

from .copier import Copier


@dataclass
class Converser(Copier, AppInteractor):
    """
    A base class for agents interacting with LLMs.
    """
    COPY_ATTRIBUTES = {'actions_and_conversations', 'conversation_name', 'assistant_agent',
                       'user_agent'}
    LLM_PARAMETERS = {}  # default parameters to pass to the LLM. e.g. {'temperature': 0.0, 'max_tokens': 30}
    actions_and_conversations: ActionsAndConversations = None

    model_engine: ModelEngine = field(default_factory=lambda: ModelEngine.DEFAULT)
    # The openai model engine to use (a call to apply_get_and_append_assistant_message can override this value).

    llm_parameters: dict[str, Any] = None

    system_prompt: str = 'You are a helpful scientist.'

    assistant_agent: Agent = None
    user_agent: Agent = None

    conversation_name: str = 'default'
    is_new_conversation: Optional[bool] = True
    # if True, we assert that no conversation with the same name exists.
    # if False, we assert that a conversation with the same name exists.
    # if None, we make sure the conversation is new by changing the name, if needed.

    driver: str = ''

    def __post_init__(self):
        conversation_exists = self.conversation_name in self.actions_and_conversations.conversations
        if self.is_new_conversation is False:
            assert conversation_exists, f'Conversation "{self.conversation_name}" does not exist.'
        elif self.is_new_conversation is True:
            assert not conversation_exists, f'Conversation "{self.conversation_name}" already exists.'
        else:
            if conversation_exists:
                self.conversation_name = self.actions_and_conversations.conversations.get_new_conversation_name(
                    self.conversation_name)
        if self.llm_parameters is None:
            self.llm_parameters = self.LLM_PARAMETERS

        self.conversation_manager = ConversationManager(
            actions_and_conversations=self.actions_and_conversations,
            conversation_name=self.conversation_name,
            driver=self.driver if self.driver is not None else type(self).__name__,
            assistant_agent=self.assistant_agent,
            user_agent=self.user_agent,
        )

    @property
    def user_skin_name(self):
        return None if self.user_agent is None else self.user_agent.skin_name

    @property
    def assistant_skin_name(self):
        return None if self.assistant_agent is None else self.assistant_agent.skin_name

    @property
    def conversation(self):
        return self.conversation_manager.conversation

    def _print_conversation_header(self):
        print_and_log_magenta('==== Starting conversation ' + '=' * (TEXT_WIDTH - 27))
        print_and_log_magenta(self.conversation_name.center(TEXT_WIDTH))
        print_and_log_magenta('=' * TEXT_WIDTH)

    def _upon_conversation_initiation(self):
        self._print_conversation_header()
        self._app_clear_panels()
        self._app_set_focus_on_panel(PanelNames.RESPONSE)  # The Prompt panel might stay empty for a while.

    def initialize_conversation_if_needed(self):
        if self.conversation is None:
            self._upon_conversation_initiation()
        self.conversation_manager.initialize_conversation_if_needed()
        if len(self.conversation) == 0 and self.system_prompt:
            self.apply_append_system_message(self.system_prompt)

    def comment(self, comment: StrOrReplacer, tag: Optional[StrOrReplacer] = None, as_action: bool = True,
                **kwargs):
        """
        Print a comment, either directly, or as an action appending a COMMENTER message to the conversation (default).
        """
        comment = format_value(self, comment)
        if as_action:
            self.conversation_manager.append_commenter_message(
                content=comment,
                tag=tag,
                **kwargs)
        else:
            print_and_log_red(comment)

    @_raise_if_reset
    def apply_get_and_append_assistant_message(self, tag: Optional[StrOrReplacer] = None,
                                               comment: Optional[StrOrReplacer] = None,
                                               is_code: bool = False, previous_code: Optional[str] = None,
                                               is_json: bool = False,
                                               model_engine: Optional[ModelEngine] = None,
                                               hidden_messages: GeneralMessageDesignation = None,
                                               expected_tokens_in_response: int = None,
                                               send_to_app: bool = True,
                                               **kwargs) -> Message:
        model_engine = model_engine or self.model_engine or ModelEngine.DEFAULT
        with self._app_temporarily_set_panel_status(PanelNames.RESPONSE,
                                                    f'Waiting for Performer LLM ({model_engine})...'):
            message = self.conversation_manager.get_and_append_assistant_message(
                tag=tag,
                comment=comment,
                is_code=is_code, previous_code=previous_code,
                is_json=is_json,
                model_engine=model_engine,
                expected_tokens_in_response=expected_tokens_in_response,
                hidden_messages=hidden_messages,
                **{**self.llm_parameters, **kwargs})
            if send_to_app and self.app:
                self._app_send_prompt(PanelNames.RESPONSE, message.pretty_content(with_header=False, is_html=True),
                                      provided_as_html=True)
        return message

    @_raise_if_reset
    def _show_and_edit_content(self, content: StrOrReplacer,
                               editing_title: str, editing_instructions: str, in_field_instructions: str,
                               send_to_app: Optional[bool], app_panel: PanelNames, sleep_for: Optional[float]) -> str:
        content = format_value(self, content)
        if send_to_app and self.app:
            if editing_title or editing_instructions:
                content = self._app_receive_text(app_panel, content,
                                                 title=editing_title, instructions=editing_instructions,
                                                 in_field_instructions=in_field_instructions, sleep_for=sleep_for)
            self._app_send_prompt(app_panel, content, from_md=True, demote_headers_by=1, sleep_for=sleep_for)
        return content

    @_raise_if_reset
    def apply_append_user_message(self, content: StrOrReplacer, tag: Optional[StrOrReplacer] = None,
                                  comment: Optional[StrOrReplacer] = None,
                                  ignore: bool = False,
                                  previous_code: Optional[str] = None, is_background: bool = False,
                                  send_to_app: Optional[bool] = None, app_panel: PanelNames = PanelNames.FEEDBACK,
                                  editing_title: str = None, editing_instructions: str = None,
                                  in_field_instructions: Optional[str] = None,
                                  sleep_for: Union[None, float, bool] = 0,
                                  **kwargs):
        if send_to_app is None:
            send_to_app = not is_background and not ignore
        is_edit = editing_title or editing_instructions
        if is_edit:
            content = \
                self._show_and_edit_content(content, editing_title, editing_instructions, in_field_instructions,
                                            send_to_app, app_panel, sleep_for)
        result = self.conversation_manager.append_user_message(
            content=content,
            tag=tag,
            comment=comment,
            ignore=ignore,
            previous_code=previous_code, is_background=is_background, **kwargs)
        if not is_edit:
            self._show_and_edit_content(content, editing_title, editing_instructions, in_field_instructions,
                                        send_to_app, app_panel, sleep_for)
        return result

    @_raise_if_reset
    def apply_append_system_message(self, content: StrOrReplacer, tag: Optional[StrOrReplacer] = None,
                                    comment: Optional[StrOrReplacer] = None,
                                    ignore: bool = False,
                                    send_to_app: Optional[bool] = None,
                                    **kwargs):
        if send_to_app is None:
            send_to_app = not ignore
        if send_to_app and self.app:
            self._app_send_prompt(PanelNames.SYSTEM_PROMPT, content)
        return self.conversation_manager.append_system_message(
            content=format_value(self, content),
            tag=tag,
            comment=comment,
            ignore=ignore,
            **kwargs)

    @_raise_if_reset
    def apply_append_surrogate_message(self, content: StrOrReplacer,
                                       tag: Optional[StrOrReplacer] = None, comment: Optional[StrOrReplacer] = None,
                                       ignore: bool = False,
                                       previous_code: Optional[str] = None, is_background: bool = False,
                                       send_to_app: Optional[bool] = False,
                                       **kwargs):
        if send_to_app is None:
            send_to_app = not is_background and not ignore
        if send_to_app and self.app:
            self._app_send_prompt(PanelNames.RESPONSE, content)
        return self.conversation_manager.append_surrogate_message(
            content=format_value(self, content),
            tag=tag,
            comment=comment,
            ignore=ignore,
            previous_code=previous_code, is_background=is_background, **kwargs)

    def apply_delete_messages(self, message_designation: GeneralMessageDesignation, comment: Optional[str] = None):
        return self.conversation_manager.delete_messages(message_designation, comment=comment)

    def set(self, **kwargs):
        """
        Set attributes of the class.
        """
        for key, value in kwargs.items():
            setattr(self, key, value)
        return self



from typing import Dict, Any, Tuple, Optional

import pandas as pd

from data_to_paper.utils.check_type import raise_on_wrong_func_argument_types_decorator
from data_to_paper.utils.dataframe import extract_df_axes_labels


def is_str_in_df(df: pd.DataFrame, s: str):
    return s in extract_df_axes_labels(df)


AbbrToNameDef = Dict[Any, Tuple[Optional[str], Optional[str]]]


@raise_on_wrong_func_argument_types_decorator
def split_mapping(abbrs_to_names_and_definitions: AbbrToNameDef):
    abbrs_to_names = {abbr: name for abbr, (name, definition) in
                      abbrs_to_names_and_definitions.items() if name is not None}
    names_to_definitions = {name or abbr: definition for abbr, (name, definition) in
                            abbrs_to_names_and_definitions.items() if definition is not None}
    return abbrs_to_names, names_to_definitions



def convert_to_weighted_edgelist(input_file, output_file, with_weights=True):
    with open(input_file, 'r') as infile, open(output_file, 'w') as outfile:
        for line in infile:
            # Separate edge data from weight attribute
            node1, node2, _, weight = line.split()
            if with_weights:
                weight = weight.split('}')[0]
                outfile.write(f"{node1} {node2} {weight}\n")
            else:
                outfile.write(f"{node1} {node2}\n")


convert_to_weighted_edgelist('congress.edgelist', 'congress_edges_weighted.dat', with_weights=True)
convert_to_weighted_edgelist('congress.edgelist', 'congress_edges.dat', with_weights=False)



from dataclasses import dataclass, field
from typing import Dict, List, Optional, Collection

from pathlib import Path

from data_to_paper.servers.custom_types import Citation
from data_to_paper.latex.latex_doc import LatexDocument

from .base_products_to_file import BaseFileProducer


@dataclass
class BaseLatexToPDF(BaseFileProducer):
    """
    Allows creating a pdf based on a LatexDocument template whose sections are populated from the Products.
    """
    COPY_ATTRIBUTES = ('latex_document',)

    latex_document: LatexDocument = field(default_factory=LatexDocument)
    paper_section_names: List[str] = None
    figures_folder: Optional[Path] = None  # folder that includes png files for the pdf

    def _get_sections(self) -> Dict[str, str]:
        """
        Return a mapping from section names to their content for all the sections of the paper.
        Should include the 'title' and 'abstract' as keys if the pdf should have a title and abstract.
        """
        return {}

    def _get_references(self) -> Collection[Citation]:
        """
        Return all the references cited.
        """
        return set()

    def _get_appendix(self) -> Optional[str]:
        """
        Return the content of the appendix.
        """
        return None

    def assemble_compile_paper(self) -> str:
        return self.latex_document.compile_document(
            content=self._get_sections(),
            appendix=self._get_appendix(),
            references=self._get_references(),
            format_cite=True,
            file_stem=self.output_file_stem,
            figures_folder=self.figures_folder,
            output_directory=str(self.output_directory),
        )[0]



from pygments.formatters.html import HtmlFormatter


# Colors:
CURRENT_STEP_COLOR = '#005599'  # darker blue
PANEL_HEADER_COLOR = CURRENT_STEP_COLOR  # "#0077cc"  # dark blue
EDIT_TEXT_COLOR = "#20b0ff"  # light gray
SUBMIT_BUTTON_COLOR = '#008000'  # dark green
BACKGROUND_COLOR = "#151515"
APP_BACKGROUND_COLOR = "#303030"

CSS = '''
* {
    color: white;
}
p {
    font-family: Arial, sans-serif;
    font-size: 14px;
    color: white;
}
.runtime_error {
    color: red;
    font-family: Consolas, 'Courier New', monospace; font-size: 14px;
}
.markdown {
    font-family: Arial, sans-serif;
    font-size: 14px;
    color: white;
    overflow-wrap: break-word; /* Allows the words to break and wrap onto the next line */
    word-wrap: break-word; /* Older syntax, similar to overflow-wrap */
    white-space: normal; /* Overrides pre to allow wrapping */
    margin-bottom: 0.5em;
}
.tripled_quote {
    font-family: Arial, sans-serif;
    font-size: 14px;
    color: #c0c0c0;
    overflow-wrap: break-word; /* Allows the words to break and wrap onto the next line */
    word-wrap: break-word; /* Older syntax, similar to overflow-wrap */
    white-space: normal; /* Overrides pre to allow wrapping */
}
.codeline {
    font-family: Consolas, 'Courier New', monospace;
}
h1 {
    color: #0066cc;
    font-size: 18px;
}
h2 {
    color: #0099cc;
    font-size: 16px;
}
h3 {
    color: #00cccc;
    font-size: 14px;
}
li {
    margin-left: 20px;
    padding-left: 0;
    list-style-type: disc;
    margin-bottom: 0.5em;
}
'''

formatter = HtmlFormatter(style="monokai")
css = formatter.get_style_defs('.highlight')
additional_css = ".highlight, .highlight pre { background: " + BACKGROUND_COLOR + "; }"

# combine the CSS with the additional CSS:
CSS += css + additional_css

SCROLLBAR_STYLE = """
QScrollBar:horizontal {
    border: none;
    background: transparent; /* Remove background color */
    height: 8px; /* Make the scrollbar smaller */
    margin: 0px 0px 0px 0px;
    border-radius: 4px; /* Adjust border-radius to match new size */
}

QScrollBar::handle:horizontal {
    background: #a0a0a0;
    min-width: 20px;
    border-radius: 4px; /* Adjust border-radius to match new size */
}

QScrollBar::add-line:horizontal, QScrollBar::sub-line:horizontal {
    border: none;
    background: transparent;
    width: 0px;
    height: 0px;
}

QScrollBar:vertical {
    border: none;
    background: transparent; /* Remove background color */
    width: 8px; /* Make the scrollbar smaller */
    margin: 0px 0px 0px 0px;
    border-radius: 4px; /* Adjust border-radius to match new size */
}

QScrollBar::handle:vertical {
    background: #a0a0a0;
    min-height: 20px;
    border-radius: 4px; /* Adjust border-radius to match new size */
}

QScrollBar::add-line:vertical, QScrollBar::sub-line:vertical {
    border: none;
    background: transparent;
    width: 0px;
    height: 0px;
}

QScrollBar::up-arrow:vertical, QScrollBar::down-arrow:vertical,
QScrollBar::left-arrow:horizontal, QScrollBar::right-arrow:horizontal {
    background: none;
    border: none;
}

QScrollBar::add-page:vertical, QScrollBar::sub-page:vertical,
QScrollBar::add-page:horizontal, QScrollBar::sub-page:horizontal {
    background: none;
}
"""

APP_STYLE = """
QMainWindow {
   background-color: black;
}
""".replace('black', APP_BACKGROUND_COLOR) + SCROLLBAR_STYLE

TABS_STYLE = """
QTabWidget::pane { /* The tab widget frame */
    border-top: 2px solid #202020;
}

QTabBar::tab {
    background-color: #303030;
    color: white;
    border: 2px solid #505050; /* Visible borders around tabs */
    border-bottom-color: #303030; /* Same as background to merge with the tab pane */
    padding: 5px; /* Spacing within the tabs */
}

QTabBar::tab:selected {
    background-color: #505050;
    border-color: #606060; /* Slightly lighter border to highlight the selected tab */
    border-bottom-color: #505050; /* Merge with the tab pane */
}

QTabBar::tab:hover {
    background-color: #404040; /* Slightly lighter to indicate hover state */
}
"""

QEDIT_STYLE = r"""
QTextEdit {
    font-family: Arial, sans-serif;
    font-size: 14px;
    color: {color};
    background-color: {background_color};
    font-size: 14px;
}
""".replace('{color}', EDIT_TEXT_COLOR).replace('{background_color}', BACKGROUND_COLOR)

HTMLPOPUP_STYLE = """
* {
    background-color: """ + BACKGROUND_COLOR + """;
    color: white;
}
QPushButton {
    background-color: #E3E0DA; 
    color: """ + BACKGROUND_COLOR + """;
}
""" + SCROLLBAR_STYLE

STEP_PANEL_BUTTON_STYLE = """
QPushButton {{
    background-color: {background_color};
    border-radius: 5px;
}}
QPushButton:pressed {{
    background-color: {pressed_color};
}}
"""

STEP_PANEL_RESET_BUTTON_STYLE = """
QPushButton {{
    background-color: {background_color};
    border-radius: 5px;
}}
QPushButton:pressed {{
    background-color: {pressed_color};
}}
"""

MAIN_SPLITTER_STYLE = """
QSplitter::handle {
    width: 1px;
    background-color: #202020;
    }
"""

QCHECKBOX_STYLE = """
QCheckBox {
    color: white;
}

QCheckBox::indicator {
    width: 15px;
    height: 15px;
}
"""



import pandas as pd

from ...analysis.my_utils.df_to_figure import analysis_df_to_figure


def displayitems_df_to_figure(df: pd.DataFrame, filename: str, **kwargs):
    """
    Same as analysis_df_to_figure, but for displayitems.
    See docstring of analysis_df_to_figure for more information.
    """
    return analysis_df_to_figure(df, filename, **kwargs)



from typing import NamedTuple, Optional


class TagPairs(NamedTuple):
    """
    end strings to extract from text.
    if right_tag=None - extract to the end.
    """
    left_tag: str
    right_tag: Optional[str]

    def __str__(self):
        if self.right_tag is None:
            return f'`{self.left_tag}`'
        else:
            return f'`{self.left_tag}` and `{self.right_tag}`'

    def wrap(self, text: str) -> str:
        """
        Wrap text with the tags.
        """
        return f'{self.left_tag}{text}{self.right_tag}'

    def is_flanking(self) -> bool:
        """
        Check if the tag is flanking.
        """
        return self.right_tag is not None


# String patterns used to wrap text for save and load. Use unique patterns, not likely to occur in conversation.
SAVE_TAGS = TagPairs('START>>>>>\n', '\n<<<<<END\n')
